"""
@file fields/discrete.py

Discrete fields (scalars or vectors) descriptions.
"""
from parmepy.constants import debug, parmesPickle
from itertools import count
from parmepy.mpi import main_rank
from parmepy.tools.timers import Timer, timed_function
import numpy as np
from parmepy.constants import ORDER, PARMES_REAL
import parmepy.tools.numpywrappers as npw
import numpy.linalg as la
from parmepy.numerics.updateGhosts import UpdateGhosts


class DiscreteField(object):
    """
    Interface for discrete fields (scalar or vector).

    A discrete field represents the data layout on each mpi process for
    a given global resolution.

    More precisely a discrete field :
    - is the discrete representation of a continuous field for some
    local resolution given by a parmepy.mpi.topology.Cartesian topology.
    - can be a vector or a scalar

    Thus, this object is the one which deals with memory allocation
    on each processor (self.data member)

    Constructor (init) process sets the associated topology and allocate
    memory for the data (self.data object, numpy arrays).

    Data are initialized to zero by default. It's possible to provide
    a user-defined function to compute another set of initial values
    by using discrete.initialize(formula=your_function_name).

    Example 1 : direct build of a discrete field
    (topo being a pre-defined parmepy.mpi.topology.Cartesian object) :
    \code
    velocity = ScalarField(topo)
    def myfunc(x,y,z):
        velx = sin(x)
        vely = cos(y)
        velz = sin(z)
        return velx, vely, velz

    velocity.initialize(formula=myfunc)
    \endcode

    Example 2 : description of a continuous and discretisation
    \code
    velocity = Field()
    # discretisation of the continuous field on two different topologies
    # i.e. creation of two discrete fields
    velocity.discretize(topo)
    velocity.discretize(topo2)
    # Initialisation of each discrete fields of velocity with
    # velocity(x,y,z) = myfunc(x,y,z) for each point of the grid
    velocity.initialize(formula=myfunc)
    \endcode
    """
    ## Counter for discrete field instanciations.
    ## Used to set default name and for serialization ids.
    __field_counter = count(0)

    @debug
    def __new__(cls, *args, **kw):
        return object.__new__(cls, *args, **kw)

    @debug
    def __init__(self, topology, isVector=False, name=None, nbComponents=None):
        """ Discretize a field on a given topology.

        @param topology : a topology (mpi topo + mesh, parmepy.mpi.topology)
        @param name : field name
        """
        ## Topology used to distribute/discretize the present field.
        self.topology = topology

        ## Id (unique) for the field
        self.__id = self.__field_counter.next()
        ## Field name.
        if name is not None:
            self.name = name
        else:
            self.name = 'unamed'
        ## Field dimension.
        self.dimension = topology.domain.dimension
        ## Field resolution.
        self.resolution = self.topology.mesh.resolution
        ## Application domain of the variable.
        self.domain = self.topology.domain
        ## Scalar or vector
        self.isVector = isVector
        ## Object to store computational times
        self.timer = Timer(self, ' (' + self.name + ')')
        ## Number of components of the field
        if nbComponents is None:
            self.nbComponents = self.dimension if self.isVector else 1
        else:
            self.nbComponents = nbComponents
        ## The memory space for data ...
        self.data = [npw.zeros(self.resolution)
                     for d in xrange(self.nbComponents)]

        ## Ghost synchronization function
        self.synchro_ghosts = self._empty_synchro_ghosts
        self._the_gosts_synchro = None
        if self.topology.hasGhosts:
            self.synchro_ghosts = self._synchro_ghosts
            self._the_ghosts_synchro = UpdateGhosts(self.topology,
                                                    self.nbComponents)

    def __getitem__(self, i):
        """ Access to the content of the field.
        @param i : requested index.
        @return component i of the field.
        """
        return self.data[i]

    def __setitem__(self, i, value):
        """
        Set the content of component i of the field.
        """
        self.data[i][...] = value

    def _empty_synchro_ghosts(self):
        pass

    def _synchro_ghosts(self):
        """Synchronize ghosts."""
        self._the_ghosts_synchro(self.data)

    @debug
    @timed_function
    def initialize(self, formula=None, doVectorize=False,
                   currentTime=0., *args):
        """
        Initialize values with a given formula.

        @param formula : a user-defined python function
        @param doVectorize : true if formula must be vectorized
        (i.e. is of type 'user_func_2', see parmepy.fields for details)
        @param doVectorize : true (default = false) if formula must
        be vectorized to handle numpy arrays. See notes about fields
        initialization in parmepy.fields.
        @param currentTime : current time (default set to 0.0)
        @param args : extra (optional) parameters
        """
        if formula is not None:
            # Argument of formula. Usually : x, y, z, t, extras
            arg_list = self.topology.mesh.coords + (currentTime,) + args
            if doVectorize:  # input formula is not defined for numpy arrays
                if isinstance(formula, np.lib.function_base.vectorize):
                    v_formula = formula
                else:
                    v_formula = np.vectorize(formula)
                if len(self.data) == 1:
                    self.data[0][...] = v_formula(*arg_list)
                elif len(self.data) == 2:
                    self.data[0][...], self.data[1][...] = v_formula(*arg_list)
                elif len(self.data) == 3:
                    self.data[0][...], self.data[1][...], self.data[2][...] = \
                        v_formula(*arg_list)
                else:
                    # Warning : in this case, self.data[i] memory will
                    # be reallocated.
                    print ("Warning : reallocation of memory for fields data\
                        during initialisation. See parmepy.fields\
                        documentation for details.")
                    self.data = v_formula(*arg_list)
                    # Ensure that data is of the right type,
                    # in the right order.
                    for i in xrange(self.nbComponents):
                        self.data[i][...] = np.asarray(self.data[i],
                                                       dtype=PARMES_REAL,
                                                       order=ORDER)

            else:
                # In that cas, we assume that formula has been properly
                # defined as a function of (res, x, y, ...),
                # res[i], x, y ... being numpy arrays.
                self.data = formula(self.data, *arg_list)
        else:
            # No formula, set all components to zero"
            for d in xrange(self.nbComponents):
                self.data[d][...] = 0.0
        assert all([(s == self.resolution).all()
                    for s in [dat.shape for dat in self.data]]),\
            "WARNING: The shape of " + self.name + " has changed during"\
            " field initialisation."

    def norm(self):
        """
        Compute a 2-norm for the discrete field
        p-norm = (sum data[d](i,...)**p)**1/p for d = 1..dim
        sum on all grid points excluding ghosts.
        Warning : this is probably not a proper way
        to compute the 'real' norm of the field but
        just a way to check field distribution. To be improved.
        """
        result = npw.zeros((self.nbComponents))
        gResult = npw.zeros((self.nbComponents))
        indNoGhost = self.topology.mesh.iCompute
        for d in range(self.nbComponents):
            result[d] = la.norm(self.data[d][indNoGhost]) ** 2
        self.topology.comm.Allreduce(result, gResult)
        return gResult ** 0.5

    def normh(self):
        """
        Compute a 'grid-norm' for the discrete field
        norm = ( hd * sum data[d](i,...)**2)**0.5 for d = 1..dim
        and hd the grid step size in direction d.
        Sum on all grid points excluding ghosts.
        """
        result = npw.zeros((self.nbComponents))
        gResult = npw.zeros((self.nbComponents))
        indNoGhost = self.topology.mesh.iCompute
        step = self.topology.mesh.space_step
        for d in range(self.nbComponents):
            result[d] = (step[d] * la.norm(self.data[d][indNoGhost])) ** 2
        self.topology.comm.Allreduce(result, gResult)
        return gResult ** 0.5

    @debug
    @timed_function
    def dump(self, filename, mode=None):
        """
        Dump (serialize) the data of the field into filename.
        @param filename : name of the file in which data are serialized
        @param mode : set mode='append' to add data to an existing file
        """
        filename += '_rk_'
        filename += str(main_rank)
        # create a new db
        if mode is None:
            db = parmesPickle(filename, mode='store')
        elif mode is 'append':
            # use an existing db
            db = parmesPickle(filename, mode='load')

        #for dim in xrange(self.dimension):
        #    idd = self.name + '_' + str(dim)
        db.dump(self.data, self.name)

    @debug
    @timed_function
    def load(self, filename, fieldname=None):
        """
        load data from a file build
        with cPickle serialization.
        @param filename : name of the file in which data are serialized
        """
        if fieldname is None:
            fieldname = self.name
        filename += '_rk_'
        filename += str(main_rank)
        db = parmesPickle(filename, mode='load')
        for dim in xrange(self.nbComponents):
            self.data[dim] = db.load(fieldname)[0][dim]

    def zero(self):
        """ set all components to zero"""
        for dim in xrange(self.nbComponents):
            self.data[dim][...] = 0.0

    def integrate_on_proc(self, box, useSlice=True, component=0):
        """
        integrate the field on a control box, on the current processus
        @param box : a parmepy.domain.obstacles.controlBox.ControlBox
        @param useSlice : true if integrate with slices else integrate with
        boolean numpy array
        @param component : component number of the field to integrate
        """
        if useSlice:
            cond = box.slices[self.topology]
        else:
            iC = self.topology.mesh.iCompute
            cond = box.ind[self.topology][0][iC]
        dvol = npw.prod(self.topology.mesh.space_step)
        result = npw.sum(self.data[component][cond])
        result *= dvol
        return result

    def integrate(self, box, useSlice=True, component=0,
                  root=0, mpiall=True):
        """
        integrate the field on a control box, on the current processus
        @param box : a parmepy.domain.obstacles.controlBox.ControlBox
        @param useSlice : true if integrate with slices else integrate with
        boolean numpy array
        @param component : component number of the field to integrate
        @param root : root process for mpi reduce
        @param mpiall : true for allreduce, else reduce on root
        """
        res = self.integrate_on_proc(box, useSlice, component)
        if mpiall:
            return self.topology.comm.allreduce(res)
        else:
            return self.topology.comm.reduce(res, root=root)

    def integrateOnSurface(self, surf, useSlice=True, component=0,
                           root=0, mpiall=True):
        """
        integrate the field on a surface, on the current processus
        @param surf : a parmepy.domain.obstacles.controlBox.planes.SubPlane
        @param useSlice : true if integrate with slices else integrate with
        boolean numpy array
        @param component : component number of the field to integrate
        @param root : root process for mpi reduce
        @param mpiall : true for allreduce, else reduce on root
        Warning : surf slices or cond must have been computed for integration
        purpose, that is last point in each dir must not be included.
        """
        res = self.integrateOnSurf_proc(surf, useSlice, component)
        if mpiall:
            return self.topology.comm.allreduce(res)
        else:
            return self.topology.comm.reduce(res, root=root)

    def integrateOnSurf_proc(self, surf, useSlice=True, component=0):
        """
        integrate the field on a surface, on the current processus
        @param surf : a parmepy.domain.obstacles.controlBox.planes.Plane
         or SubPlane
        @param useSlice : true if integrate with slices else integrate with
        boolean numpy array
        @param component : component number of the field to integrate
        Warning : surf slices or cond must have been computed for integration
        purpose, that is last point in each dir must not be included.
        """
        if useSlice:
            cond = surf.slices[self.topology]
        else:
            iC = self.topology.mesh.iCompute
            cond = surf.ind[self.topology][0][iC]
        dirs = np.where(surf.normal == 0)[0]
        dS = npw.prod(self.topology.mesh.space_step[dirs])
        result = npw.sum(self.data[component][cond])
        result *= dS
        return result
