"""
Testing parmepy.field.continuous.Field
"""
from parmepy.fields.continuous import Field
from parmepy.domain.box import Box
from parmepy.mpi.topology import Cartesian
from parmepy.domain.obstacle.planes import SubPlane
import sys
import numpy as np


def test_continuous():
    """ Basic continuous field construction """

    print (sys.path)
    dom = Box()
    cf = Field(dom)
    assert cf.isVector is False
    assert cf.domain is dom
    cf = Field(dom, isVector=True)
    assert cf.isVector is True
    assert cf.domain is dom


def test_analytical():
    """Test formula"""
    dom = Box()
    formula = lambda x, y, z: (x * y * z, 1.0, -x * y * z)
    caf = Field(dom, formula, doVectorize=True)
    assert caf.value(0., 0., 0.) == (0., 1., 0.)
    assert caf.value(1., 1., 1.) == (1., 1., -1.)
    assert caf.value(1., 2., 3.) == (6., 1., -6.)


def test_analytical_reset():
    """Test formula"""
    dom = Box()
    formula = lambda x, y, z: (x * y * z, 1.0, -x * y * z)
    caf = Field(dom, formula, doVectorize=True)
    assert caf.value(0., 0., 0.) == (0., 1., 0.)
    assert caf.value(1., 1., 1.) == (1., 1., -1.)
    assert caf.value(1., 2., 3.) == (6., 1., -6.)
    form2 = lambda x, y, z, dt: (x * y * z, 1.0, -x * y * z * dt)
    caf.setFormula(form2, doVectorize=True)
    assert caf.value(0., 0., 0., 0.) == (0., 1., 0.)
    assert caf.value(1., 1., 1., 1.) == (1., 1., -1.)
    assert caf.value(1., 2., 3., 4.) == (6., 1., -24.)


def test_discretization():
    """Test multiple discretizations"""
    dom = Box()
    csf = Field(dom)
    cvf = Field(dom, isVector=True)
    resolTopo = [33, 33, 17]
    topo = Cartesian(dom, 3, resolTopo)
    csf.discretize(topo)
    cvf.discretize(topo)
    assert np.equal(csf.discreteFields[topo].resolution,
                    csf.discreteFields[topo].data[0].shape).all()
    assert np.equal(cvf.discreteFields[topo].resolution,
                    cvf.discreteFields[topo].data[0].shape).all()
    assert np.equal(cvf.discreteFields[topo].resolution,
                    cvf.discreteFields[topo].data[1].shape).all()
    assert np.equal(cvf.discreteFields[topo].resolution,
                    cvf.discreteFields[topo].data[2].shape).all()


def test_integrate_onSurf():
    dom = Box()
    myf = Field(dom, isVector=True)
    resolTopo = [33, 33, 17]
    topo = Cartesian(dom, 3, resolTopo)
    fdiscr = myf.discretize(topo)
    fdiscr[0][...] = 1.0
    normal = [1, 0, 0]
    hh = topo.mesh.space_step
    surf = SubPlane(domain=dom, normal=normal, point=dom.origin,
                    lengths=dom.length - hh,
                    epsilon=topo.mesh.space_step[0]/2.)

    surf.discretize(topo)
    res = myf.integrateOnSurface(surf, topo)
    sref = dom.length[1] * dom.length[2]
    assert(abs(res - sref) < 1e-6)

# This may be useful to run mpi tests
if __name__ == "__main__":
    test_continuous()
    test_analytical()
    test_analytical_reset()
    test_integrate_onSurf()
    test_discretization()
