"""
@file gpu_particle_advection.py

Discrete advection representation
"""
from abc import ABCMeta, abstractmethod
from parmepy import __VERBOSE__
from parmepy.constants import np, debug, PARMES_INDEX, S_DIR, \
    PARMES_REAL, PARMES_INTEGER
from parmepy.methods_keys import TimeIntegrator, Interpolation, Remesh, \
    Support, Splitting, Precision
from parmepy.numerics.integrators.runge_kutta2 import RK2
from parmepy.numerics.interpolation import Linear
from parmepy.numerics.remeshing import L2_1
from parmepy.operator.discrete.particle_advection import ParticleAdvection
from parmepy.gpu import cl
from parmepy.gpu.tools import get_opencl_environment
from parmepy.gpu.gpu_kernel import KernelLauncher
from parmepy.tools.timers import Timer
from parmepy.operator.discrete.discrete import DiscreteOperator
from parmepy.mpi import MPI


class GPUParticleAdvection(ParticleAdvection):
    """
    Particle advection operator representation on GPU.

    """
    __metaclass__ = ABCMeta

    @debug
    @abstractmethod
    def __init__(self, velocity, advectedFields, d,
                 part_position=None, part_advectedFields=None,
                 platform_id=None, device_id=None,
                 device_type=None,
                 method=None,
                 src=None, batch_nb=None,
                 isMultiScale=False):
        """
        Create a Advection operator.
        Work on a given field (scalar or vector) at a given velocity to compute
        advected values.
        OpenCL kernels are build once per dimension in order to handle
        directional splitting with resolution non uniform in directions.

        @param velocity : Velocity field
        @param advectedFields : Advected fields
        @param d : Direction to advect
        @param platform_id : OpenCL platform id (default = 0).
        @param device_id : OpenCL device id (default = 0).
        @param device_type : OpenCL device type (default = 'gpu').
        @param method : the method to use. {'m4prime', 'm6prime', 'm8prime',
        'l6star'}
        @param src : User OpenCL sources.
        @param splittingConfig : Directional splitting configuration
        (parmepy.numerics.splitting.Splitting.__init__)
        """
        if method is None:
            method = {TimeIntegrator: RK2,
                      Interpolation: Linear,
                      Remesh: L2_1,
                      Support: 'gpu_2k',
                      Splitting: 'o2',
                      Precision: PARMES_REAL}
        ParticleAdvection.__init__(self, velocity, advectedFields, d,
                                   part_position, part_advectedFields, method,
                                   isMultiScale=isMultiScale)
        self.method = method
        if not Precision in self.method.keys():
            self.method[Precision]=PARMES_REAL
        self.user_gpu_src = src
        self.num_method = None
        self.dim = self.advectedFields[0].dimension
        self.cl_env = get_opencl_environment(
            platform_id=platform_id, device_id=device_id,
            device_type=device_type, precision=self.method[Precision],
            comm=self.advectedFields[0].topology.comm)
        self._main_rank = self.advectedFields[0].topology.comm.Get_rank()
        self._main_size = self.advectedFields[0].topology.comm.Get_size()

        resol = self.advectedFields[0].topology.mesh.resolution
        v_resol = self.velocity.topology.mesh.resolution
        shape = np.ones((self.dim,), dtype=PARMES_INTEGER)
        shape[:self.dim] = resol[...]
        v_shape = np.ones((self.dim,), dtype=PARMES_INTEGER)
        v_shape[:self.dim] = v_resol[...]
        self.batch_nb = np.ones((self.dim,), dtype=PARMES_INTEGER)
        if self.dir == self.dim - 1:
            batch_dir = 0
        else:
            batch_dir = self.dim - 1
        gh_velo = self.velocity.topology.ghosts
        gh_adv = self.advectedFields[0].topology.ghosts
        if batch_nb is None:
            # Automatic computation of batch number regarding available mem
            # Ghosts are not divided by 2: shape = (shape - 2*gh)/2 + 2*gh
            prev_mem_use = self.cl_env.gpu_comm.allreduce(
                self.globalMemoryUsagePreview(v_shape, shape))
            while prev_mem_use >= self.cl_env.available_mem:
                shape[batch_dir] /= 2
                shape[batch_dir] += gh_adv[batch_dir]
                v_shape[batch_dir] /= 2
                v_shape[batch_dir] += gh_velo[batch_dir]
                self.batch_nb[batch_dir] *= 2
                prev_mem_use = self.cl_env.gpu_comm.allreduce(
                    self.globalMemoryUsagePreview(v_shape, shape))
        else:
            # User batch number
            shape[batch_dir] /= batch_nb
            shape[batch_dir] += gh_adv[batch_dir]
            v_shape[batch_dir] /= batch_nb
            v_shape[batch_dir] += gh_velo[batch_dir]
            self.batch_nb[batch_dir] = batch_nb
        if np.prod(self.batch_nb) > 1 and self._main_rank == 0:
            print "Using batch for variables:", self.batch_nb

        # Functions to get the appropriate vectors for the current direction
        self._reorderVect = lambda v: v
        if self.dim == 2 and self.dir == 1:
            self._reorderVect = lambda v: (v[1], v[0])
        if self.dim == 3 and self.dir == 1:
            self._reorderVect = lambda v: (v[1], v[0], v[2])
        if self.dim == 3 and self.dir == 2:
            if self._main_size == 1 and self.method[Splitting].find('o2') >= 0:
                self._reorderVect = lambda v: (v[2], v[0], v[1])
            else:
                self._reorderVect = lambda v: (v[2], v[1], v[0])
        # Compute resolutions for kernels for each direction.
        ## Resolution of the local mesh
        self.resol = np.ones((3,), dtype=PARMES_INDEX)
        self.v_resol = np.ones((3,), dtype=PARMES_INDEX)
        ## Resolution of the local mesh but reoganized redarding
        ## splitting direction:
        ## direction X : XYZ
        ## direction Y : YXZ
        ## direction Z : ZYX in parallel, ZXY in sequentiel.
        self.resol_dir = np.ones((3,), dtype=PARMES_INDEX)
        self.v_resol_dir = np.ones((3,), dtype=PARMES_INDEX)
        self.resol[:self.dim] = shape[...]
        self.resol_dir[:self.dim] = self._reorderVect(shape)
        self.v_resol[:self.dim] = v_shape[...]
        self.v_resol_dir[:self.dim] = self._reorderVect(v_shape)

        # Size constants for local mesh size
        self._size_constants = ""
        for i in xrange(3):
            self._size_constants += " -D NB" + S_DIR[i]
            self._size_constants += "=" + str(self.resol[i])
        for i in xrange(3):
            self._size_constants += " -D V_NB" + S_DIR[i]
            self._size_constants += "=" + str(self.v_resol[i])
        self._size_constants += " -D V_GHOSTS_NB="
        self._size_constants += str(self.velocity.topology.ghosts[0])
        # Direction dependant constants
        self._constants = [
            " -D NB_I=NB_X -D NB_II=NB_Y -D NB_III=NB_Z " +
            "-D V_NB_I=V_NB_X -D V_NB_II=V_NB_Y -D V_NB_III=V_NB_Z",
            " -D NB_I=NB_Y -D NB_II=NB_X -D NB_III=NB_Z " +
            "-D V_NB_I=V_NB_Y -D V_NB_II=V_NB_X -D V_NB_III=V_NB_Z",
            " -D NB_I=NB_Z -D NB_II=NB_Y -D NB_III=NB_X " +
            "-D V_NB_I=V_NB_Z -D V_NB_II=V_NB_Y -D V_NB_III=V_NB_X"]
        if self._main_size == 1 and self.method[Splitting].find('o2') >= 0:
            self._constants[2] = \
                " -D NB_I=NB_Z -D NB_II=NB_X -D NB_III=NB_Y " + \
                "-D V_NB_I=V_NB_Z -D V_NB_II=V_NB_X -D V_NB_III=V_NB_Y"

        self.prg = None
        ## Object to store computational times of OpenCL kernels
        self.kernels_timer = Timer(self)
        self._num_locMem = None
        precision = self.method[Precision]
        self._kernel_cfg = self.cl_env.kernels_config[self.dim][precision]
        self.copy = None
        self.transpose_xy, self.transpose_xy_r = None, None
        self.transpose_xz, self.transpose_xz_r = None, None
        ## Global memory allocated on gpu by this operator
        self.size_global_alloc = 0
        ## Local memory allocated on gpu by this operator
        self.size_local_alloc = 0

    @debug
    def setUp(self):
        """
        Set up.
        Compute OpenCL work-item number and space index.\n
        Compile OpenCL sources.
        Allocates OpenCL buffers. Buffers are initialized with OpenCL kernels
        if named as follows : <code>init"FieldName"</code> with
        <code>"FieldName"</code> the field
        name as given by user.
        """
        self.gpu_precision = self.cl_env.precision
        self.coord_min = np.ones(4, dtype=self.gpu_precision)
        self.mesh_size = np.ones(4, dtype=self.gpu_precision)
        self.v_mesh_size = np.ones(4, dtype=self.gpu_precision)
        self.coord_min[:self.dim] = np.asarray(
            self.advectedFields[0].topology.mesh.origin,
            dtype=self.gpu_precision)
        self.mesh_size[:self.dim] = self._reorderVect(np.asarray(
            self.advectedFields[0].topology.mesh.space_step,
            dtype=self.gpu_precision))
        self.v_mesh_size[:self.dim] = self._reorderVect(np.asarray(
            self.velocity.topology.mesh.space_step,
            dtype=self.gpu_precision))
        self.build_options = ""
        self._collect_usr_cl_src(self.user_gpu_src)
        self._collect_kernels_cl_src_copy()
        self._collect_kernels_cl_src_transpositions_xy()
        if self.dim == 3:
            self._collect_kernels_cl_src_transpositions_xz()
        self._collect_kernels_cl_src()
        if __VERBOSE__:
            print "=== OpenCL Buffer allocations ==="
        self._buffer_allocations()
        if __VERBOSE__:
            print "===\n"
            print "=== OpenCL Buffer initialisation ==="
        if self.dir == 0:
            self._buffer_initialisations()

        # Beanching the proper _compute function
        if self.part_advectedFields[0].nbComponents == 1:
            self._compute = self._compute_1c
        elif self.part_advectedFields[0].nbComponents == 2:
            self._compute = self._compute_2c
        elif self.part_advectedFields[0].nbComponents == 3:
            self._compute = self._compute_3c

        # Build execution list regarding splitting:
        # In sequential configuration, data can stay in their direction
        # dependant layout between directions. In parallel, data must be
        # redistributed (between directions) in XYZ layout.
        # Splitting Strang 2nd order:
        #   3D: X(dt/2), Y(dt/2), Z(dt), Y(dt/2), X(dt/2)
        #   2D: X(dt/2), Y(dt), X(dt/2)
        if np.prod(self.batch_nb) == 1 and self.method[Splitting] == 'o2':
            if self._main_size > 1:
                if self.dim == 2:
                    self.exec_list = [
                        [self._init_copy, self._compute],  # X(dt/2)
                        [self._init_transpose_xy, self._compute,  # Y(dt)
                         self._init_transpose_xy_r, self._init_copy_r],
                        [self._init_copy, self._compute]  # X(dt/2)
                        ]
                elif self.dim == 3:
                    self.exec_list = [
                        [self._init_copy, self._compute],  # X(dt/2)
                        [self._init_transpose_xy, self._compute,  # Y(dt/2)
                         self._init_transpose_xy_r, self._init_copy_r],
                        [self._init_transpose_xz, self._compute,  # Z(dt)
                         self._init_transpose_xz_r, self._init_copy_r],
                        [self._init_transpose_xy, self._compute,  # Y(dt/2)
                         self._init_transpose_xy_r, self._init_copy_r],
                        [self._init_copy, self._compute]  # X(dt/2)
                        ]
            else:
                if self.dim == 2:
                    self.exec_list = [
                        [self._init_copy, self._compute],  # X(dt/2)
                        [self._init_transpose_xy, self._compute],  # Y(dt)
                        [self._init_transpose_xy, self._compute]  # X(dt/2)
                        ]
                elif self.dim == 3:
                    self.exec_list = [
                        [self._init_copy, self._compute],  # X(dt/2)
                        [self._init_transpose_xy, self._compute],  # Y(dt/2)
                        [self._init_transpose_xz, self._compute],  # Z(dt)
                        [self._init_transpose_xz, self._compute],  # Y(dt/2)
                        [self._init_transpose_xy, self._compute]  # X(dt/2)
                        ]

        # Splitting Strang 2nd order (fullHalf):
        #   X(dt/2), Y(dt/2), Z(dt/2), Z(dt/2), Y(dt/2), X(dt/2)
        elif np.prod(self.batch_nb) == 1 and \
                self.method[Splitting] == 'o2_FullHalf':
            if self._main_size > 1:
                if self.dim == 2:
                    self.exec_list = [
                        [self._init_copy, self._compute],  # X(dt/2)
                        [self._init_transpose_xy, self._compute],  # Y(dt/2)
                        [self._init_copy, self._compute,  # Y(dt/2)
                         self._init_transpose_xy_r, self._init_copy_r],
                        [self._init_copy, self._compute]  # X(dt/2)
                        ]
                elif self.dim == 3:
                    self.exec_list = [
                        [self._init_copy, self._compute],  # X(dt/2)
                        [self._init_transpose_xy, self._compute,  # Y(dt/2)
                         self._init_transpose_xy_r, self._init_copy_r],
                        [self._init_transpose_xz, self._compute],  # Z(dt/2)
                        [self._init_copy, self._compute,  # Z(dt/2)
                         self._init_transpose_xz_r, self._init_copy_r],
                        [self._init_transpose_xy, self._compute,  # Y(dt/2)
                         self._init_transpose_xy_r, self._init_copy_r],
                        [self._init_copy, self._compute]  # X(dt/2)
                        ]
            else:
                if self.dim == 2:
                    self.exec_list = [
                        [self._init_copy, self._compute],  # X(dt/2)
                        [self._init_transpose_xy, self._compute],  # Y(dt)
                        [self._init_copy, self._compute],  # Y(dt)
                        [self._init_transpose_xy, self._compute]  # X(dt/2)
                        ]
                elif self.dim == 3:
                    self.exec_list = [
                        [self._init_copy, self._compute],  # X(dt/2)
                        [self._init_transpose_xy, self._compute],  # Y(dt/2)
                        [self._init_transpose_xz, self._compute],  # Z(dt/2)
                        [self._init_copy, self._compute],  # Z(dt/2)
                        [self._init_transpose_xz, self._compute],  # Y(dt/2)
                        [self._init_transpose_xy, self._compute]  # X(dt/2)
                        ]
        # If batch, data are always transfered from/to host for each splitting
        # step (memory layout is applied within transfering data)
        elif np.prod(self.batch_nb) > 1 and \
                self.method[Splitting] == 'o2':
            if self.dim == 2:
                self.exec_list = [
                    [self._init_copy, self._compute],  # X(dt/2)
                    [self._init_copy, self._compute],  # Y(dt)
                    [self._init_copy, self._compute],  # X(dt/2)
                    ]
            elif self.dim == 3:
                self.exec_list = [
                    [self._init_copy, self._compute],  # X(dt/2)
                    [self._init_copy, self._compute],  # Y(dt/2)
                    [self._init_copy, self._compute],  # Z(dt)
                    [self._init_copy, self._compute],  # Y(dt/2)
                    [self._init_copy, self._compute],  # X(dt/2)
                    ]
        elif np.prod(self.batch_nb) > 1 and \
                self.method[Splitting] == 'o2_FullHalf':
            if self.dim == 2:
                self.exec_list = [
                    [self._init_copy, self._compute],  # X(dt/2)
                    [self._init_copy, self._compute],  # Y(dt/2)
                    [self._init_copy, self._compute],  # Y(dt/2)
                    [self._init_copy, self._compute],  # X(dt/2)
                    ]
            elif self.dim == 3:
                self.exec_list = [
                    [self._init_copy, self._compute],  # X(dt/2)
                    [self._init_copy, self._compute],  # Y(dt/2)
                    [self._init_copy, self._compute],  # Z(dt/2)
                    [self._init_copy, self._compute],  # Z(dt/2)
                    [self._init_copy, self._compute],  # Y(dt/2)
                    [self._init_copy, self._compute],  # X(dt/2)
                    ]
        else:
            raise ValueError('Not yet implemeted Splitting on GPU : ' +
                             self.method[Splitting])

        if np.prod(self.batch_nb) > 1:
            self._the_apply = self._apply_batch
        else:
            self._the_apply = self._apply_noBatch

        if __VERBOSE__:
            print "===\n"

    def _collect_kernels_cl_src(self):
        pass

    def _buffer_allocations(self):
        raise ValueError('This method must be implemented in subclasses.')

    def _compute_1c(self):
        raise ValueError('This method must be implemented in subclasses.')

    def _compute_2c(self):
        raise ValueError('This method must be implemented in subclasses.')

    def _compute_3c(self):
        raise ValueError('This method must be implemented in subclasses.')

    def _buffer_initialisations(self):
        """
        OpenCL buffer initializations from user OpenCL kernels.
        Looking for kernels named <code>init<FieldName></code>.
        """
        for gpudf in self.variables:
            match = 'init' + gpudf.name
            # Looking for initKernel
            if not self.prg is None:
                for k in self.prg.all_kernels():
                    k_name = k.get_info(cl.kernel_info.FUNCTION_NAME)
                    if match.find(k_name) >= 0:
                        if __VERBOSE__:
                            print gpudf.name, '-> OpenCL Kernel', k_name
                        if gpudf == self.velocity:
                            workItemNumber, gwi, lwi = \
                                self.cl_env.get_WorkItems(self.v_resol_dir)
                        else:
                            workItemNumber, gwi, lwi = \
                                self.cl_env.get_WorkItems(self.resol_dir)
                        gpudf.setInitializationKernel(KernelLauncher(
                            cl.Kernel(self.prg, k_name), self.cl_env.queue,
                            gwi, lwi))

    def _collect_kernels_cl_src_copy(self):
        """
        Compile OpenCL sources for copy kernel.
        """
        build_options = self.build_options
        # copy settings
        src, t_dim, b_rows, vec, f_space = self._kernel_cfg['copy']
        while t_dim > self.resol_dir[0] or (self.resol_dir[0] % t_dim) > 0:
            t_dim /= 2
        gwi, lwi = f_space(self.resol_dir, t_dim, b_rows, vec)

        # Build code
        build_options += " -D TILE_DIM_COPY={0}".format(t_dim)
        build_options += " -D BLOCK_ROWS_COPY={0}".format(b_rows)
        build_options += self._size_constants + self._constants[self.dir]
        prg = self.cl_env.build_src(
            src,
            build_options,
            vec)
        self.copy = KernelLauncher(prg.copy,
                                   self.cl_env.queue, gwi, lwi)

    def _collect_kernels_cl_src_transpositions_xy(self):
        """Compile OpenCL sources for transpositions kernel.
        @remark : Transpositions kernels are launched as initialization.
        Arrays are taken in their destination layout (for initialize in Y
        directions, either we came from X or Z but shapes are the Y ones).
        """
        resol = self.advectedFields[0].topology.mesh.resolution
        resol_tmp = np.empty_like(resol)

        # XY transposition settings
        is_XY_needed = self.dir == 1 or (self.dir == 0 and self._main_size == 1)
        if is_XY_needed:
            resol_tmp[...] = resol[...]
            if self.dir == 1:  # (XY -> YX)
                resol_tmp[0] = resol[1]
                resol_tmp[1] = resol[0]
                ocl_cte = " -D NB_I=NB_Y -D NB_II=NB_X -D NB_III=NB_Z"
            elif self.dir == 0:  # (YX -> XY) only for sequential
                ocl_cte = " -D NB_I=NB_X -D NB_II=NB_Y -D NB_III=NB_Z"
            build_options = self.build_options + self._size_constants
            src, t_dim, b_rows, is_padding, vec, f_space = \
                self._kernel_cfg['transpose_xy']
            while t_dim > resol_tmp[0] or t_dim > resol_tmp[1] or \
                    (resol_tmp[0] % t_dim) > 0 or (resol_tmp[1] % t_dim) > 0:
                t_dim /= 2
            gwi, lwi = f_space(resol_tmp, t_dim, b_rows, vec)

            if is_padding:
                build_options += " -D PADDING_XY=1"
            else:
                build_options += " -D PADDING_XY=0"
            build_options += " -D TILE_DIM_XY={0}".format(t_dim)
            build_options += " -D BLOCK_ROWS_XY={0}".format(b_rows)
            build_options += ocl_cte
            prg = self.cl_env.build_src(
                src,
                build_options, vec)

            self.transpose_xy = KernelLauncher(
                prg.transpose_xy, self.cl_env.queue, gwi, lwi)

        is_XY_r_needed = self.dir == 1 and self._main_size > 1
        if is_XY_r_needed:
            # Reversed XY transposition settings (YX -> XY), only in parallel
            resol_tmp[...] = resol[...]
            ocl_cte = " -D NB_I=NB_X -D NB_II=NB_Y -D NB_III=NB_Z"

            build_options = self.build_options + self._size_constants
            src, t_dim, b_rows, is_padding, vec, f_space = \
                self._kernel_cfg['transpose_xy']
            while t_dim > resol_tmp[0] or t_dim > resol_tmp[1] or \
                    (resol_tmp[0] % t_dim) > 0 or (resol_tmp[1] % t_dim) > 0:
                t_dim /= 2
            gwi, lwi = f_space(resol_tmp, t_dim, b_rows, vec)

            if is_padding:
                build_options += " -D PADDING_XY=1"
            else:
                build_options += " -D PADDING_XY=0"
            build_options += " -D TILE_DIM_XY={0}".format(t_dim)
            build_options += " -D BLOCK_ROWS_XY={0}".format(b_rows)
            build_options += ocl_cte
            prg = self.cl_env.build_src(
                src,
                build_options, vec)

            self.transpose_xy_r = KernelLauncher(
                prg.transpose_xy, self.cl_env.queue, gwi, lwi)

    def _collect_kernels_cl_src_transpositions_xz(self):
        resol = self.advectedFields[0].topology.mesh.resolution
        resol_tmp = np.empty_like(resol)

        is_XZ_needed = self.dir == 2 or (self.dir == 1 and self._main_size == 1)
        # XZ transposition settings
        if is_XZ_needed:
            resol_tmp[...] = resol[...]
            if self.dir == 1:  # ZXY -> YXZ (only for seqential)
                resol_tmp[0] = resol[1]
                resol_tmp[1] = resol[0]
                resol_tmp[2] = resol[2]
                ocl_cte = " -D NB_I=NB_Y -D NB_II=NB_X -D NB_III=NB_Z"
            elif self.dir == 2:
                if self._main_size == 1:  # YXZ -> ZXY
                    resol_tmp[0] = resol[2]
                    resol_tmp[1] = resol[0]
                    resol_tmp[2] = resol[1]
                    ocl_cte = " -D NB_I=NB_Z -D NB_II=NB_X -D NB_III=NB_Y"
                else:  # XYZ -> ZYX
                    resol_tmp[0] = resol[2]
                    resol_tmp[1] = resol[1]
                    resol_tmp[2] = resol[0]
                    ocl_cte = " -D NB_I=NB_Z -D NB_II=NB_Y -D NB_III=NB_X"

            build_options = self.build_options + self._size_constants
            src, t_dim, b_rows, b_deph, is_padding, vec, f_space = \
                self._kernel_cfg['transpose_xz']

            while t_dim > resol_tmp[0] or t_dim > resol_tmp[2] or \
                    (resol_tmp[0] % t_dim) > 0 or (resol_tmp[2] % t_dim) > 0:
                t_dim /= 2
            gwi, lwi = f_space(resol_tmp, t_dim, b_rows, b_deph, vec)
            if is_padding:
                build_options += " -D PADDING_XZ=1"
            else:
                build_options += " -D PADDING_XZ=0"
            build_options += " -D TILE_DIM_XZ={0}".format(t_dim)
            build_options += " -D BLOCK_ROWS_XZ={0}".format(b_rows)
            build_options += " -D BLOCK_DEPH_XZ={0}".format(b_deph)
            build_options += ocl_cte
            prg = self.cl_env.build_src(
                src,
                build_options,
                vec)
            self.transpose_xz = KernelLauncher(
                prg.transpose_xz, self.cl_env.queue, gwi, lwi)

        is_XZ_r_needed = self.dir == 2 and self._main_size > 1
        if is_XZ_r_needed:
            # Reversed XZ transposition settings (ZYX -> XYZ)
            resol_tmp[...] = resol[...]
            ocl_cte = " -D NB_I=NB_X -D NB_II=NB_Y -D NB_III=NB_Z"
            build_options = self.build_options + self._size_constants
            src, t_dim, b_rows, b_deph, is_padding, vec, f_space = \
                self._kernel_cfg['transpose_xz']

            while t_dim > resol_tmp[0] or t_dim > resol_tmp[2] or \
                    (resol_tmp[0] % t_dim) > 0 or (resol_tmp[2] % t_dim) > 0:
                t_dim /= 2
            gwi, lwi = f_space(resol_tmp, t_dim, b_rows, b_deph, vec)
            if is_padding:
                build_options += " -D PADDING_XZ=1"
            else:
                build_options += " -D PADDING_XZ=0"
            build_options += " -D TILE_DIM_XZ={0}".format(t_dim)
            build_options += " -D BLOCK_ROWS_XZ={0}".format(b_rows)
            build_options += " -D BLOCK_DEPH_XZ={0}".format(b_deph)
            build_options += ocl_cte
            prg = self.cl_env.build_src(
                src,
                build_options,
                vec)
            self.transpose_xz_r = KernelLauncher(
                prg.transpose_xz, self.cl_env.queue, gwi, lwi)

    def _collect_usr_cl_src(self, usr_src):
        """
        Build user sources.

        """
        build_options = self.build_options + self._size_constants
        workItemNb, gwi, lwi = self.cl_env.get_WorkItems(self.resol_dir)
        v_workItemNb, gwi, lwi = self.cl_env.get_WorkItems(self.v_resol_dir)
        build_options += " -D WI_NB=" + str(workItemNb)
        build_options += " -D V_WI_NB=" + str(v_workItemNb)
        if usr_src is not None:
            self.prg = self.cl_env.build_src(usr_src, build_options)

    def _apply_noBatch(self, simulation, dtCoeff, split_id, old_dir):
        # Call OpenCL kernels
        for exe in self.exec_list[split_id]:
            exe(simulation, dtCoeff, split_id, old_dir)

    def _apply_batch(self, simulation, dtCoeff, split_id, old_dir):
        if split_id > self.dir:
            ## Returning from Z -> X : Reverse batch order
            batch_order = xrange(np.prod(self.batch_nb)-1, -1, -1)
        else:
            batch_order = xrange(np.prod(self.batch_nb))
        for b in batch_order:
            # Checking if velocity is already on device (transfered by an
            # explicit redistribute operator)
            toDevice_useless = \
                self.velocity.batch_on_device[self.dir] == b \
                and self.velocity.batch_is_on_device[self.dir]
            if not toDevice_useless:
                self.velocity.toDevice(component=self.dir, batch_d=b)

            # Checking if advected field is already on device
            toDevice_useless = np.all(
                [b_id == b and is_b for b_id, is_b
                 in zip(self.advectedFields[0].batch_on_device,
                        self.advectedFields[0].batch_is_on_device)])
            # For full-half splitting, for the second advection in Z
            # direction, the last batch is already on device (batches are
            # handled in reversed order)
            isFullHalf = (self.dim == 3 and len(self.exec_list) == 6) or \
                (self.dim == 2 and len(self.exec_list) == 4)
            isFirstLastDir = \
                (self.dim == 3 and self.dir == 2 and split_id == 2) or \
                (self.dim == 2 and self.dir == 1 and split_id == 1)
            isSecondLastDir = \
                (self.dim == 3 and self.dir == 2 and split_id == 3) or \
                (self.dim == 2 and self.dir == 1 and split_id == 2)
            isLastBatch = b == (np.prod(self.batch_nb) - 1)
            if not (isFullHalf and isSecondLastDir and isLastBatch):
                 if not toDevice_useless:
                     self.advectedFields[0].toDevice(
                         layoutDir=self.dir, batch_d=b)

            # Call OpenCL kernels
            for exe in self.exec_list[split_id]:
                exe(simulation, dtCoeff, split_id, old_dir)

            # Send back the batch to host
            # For full-half splitting, for the first advection in Z
            # direction, the last batch not get back to host, it will be
            # used in place in the second step
            if not (isFullHalf and isFirstLastDir and isLastBatch):
                self.advectedFields[0].toHost(
                    layoutDir=self.dir, batch_d=b)

    @debug
    def apply(self, simulation, dtCoeff, split_id, old_dir):
        """
        Apply operator along specified splitting direction.
        @param t : Current time
        @param dt : Time step
        @param d : Splitting direction
        @param split_id : Splitting step id
        """
        # Calling for requirements completion
        DiscreteOperator.apply(self, simulation)
        ctime = MPI.Wtime()

        # If first direction of advection, wait for work gpu fields
        # It avoid wait_for lists to increase indelinitely
        # In practice, all events are terminated so wait() resets events list
        if split_id == 0:
            for v in self._work + self.advectedFields + [self.velocity]:
                v.clean_events()
        self._the_apply(simulation, dtCoeff, split_id, old_dir)
        self._apply_timer.append_time(MPI.Wtime() - ctime)

    def _init_copy(self, simulation, dtCoeff, split_id, old_dir):
        wait_evt = self.advectedFields[0].events + \
            self.part_advectedFields[0].events
        for p, g in zip(self.advectedFields[0].gpu_data,
                        self.part_advectedFields[0].gpu_data):
            evt = self.copy(p.data, g.data, wait_for=wait_evt)
            self.part_advectedFields[0].events.append(evt)

    def _init_copy_r(self, simulation, dtCoeff, split_id, old_dir):
        wait_evt = self.advectedFields[0].events + \
            self.part_advectedFields[0].events
        for p, g in zip(self.advectedFields[0].gpu_data,
                        self.part_advectedFields[0].gpu_data):
            evt = self.copy(g.data, p.data, wait_for=wait_evt)
            self.advectedFields[0].events.append(evt)

    def _init_transpose_xy(self, simulation, dtCoeff, split_id, old_dir):
        wait_evt = self.advectedFields[0].events + \
            self.part_advectedFields[0].events
        for p, g in zip(self.advectedFields[0].gpu_data,
                        self.part_advectedFields[0].gpu_data):
            evt = self.transpose_xy(p.data, g.data, wait_for=wait_evt)
            self.part_advectedFields[0].events.append(evt)

    def _init_transpose_xy_r(self, simulation, dtCoeff, split_id, old_dir):
        wait_evt = self.advectedFields[0].events + \
            self.part_advectedFields[0].events
        for p, g in zip(self.advectedFields[0].gpu_data,
                        self.part_advectedFields[0].gpu_data):
            evt = self.transpose_xy_r(p.data, g.data, wait_for=wait_evt)
            self.part_advectedFields[0].events.append(evt)

    def _init_transpose_xz(self, simulation, dtCoeff, split_id, old_dir):
        wait_evt = self.advectedFields[0].events + \
            self.part_advectedFields[0].events
        for p, g in zip(self.advectedFields[0].gpu_data,
                        self.part_advectedFields[0].gpu_data):
            evt = self.transpose_xz(p.data, g.data, wait_for=wait_evt)
            self.part_advectedFields[0].events.append(evt)

    def _init_transpose_xz_r(self, simulation, dtCoeff, split_id, old_dir):
        wait_evt = self.advectedFields[0].events + \
            self.part_advectedFields[0].events
        for p, g in zip(self.advectedFields[0].gpu_data,
                        self.part_advectedFields[0].gpu_data):
            evt = self.transpose_xz_r(p.data, g.data, wait_for=wait_evt)
            self.part_advectedFields[0].events.append(evt)

    @debug
    def finalize(self):
        """
        Free OpenCL memory.
        """
        for k in [self.copy, self.transpose_xy, self.transpose_xy_r,
                  self.transpose_xz, self.transpose_xz_r]:
            if k is not None:
                for f_timer in k.f_timer:
                    self.kernels_timer.addFunctionTimer(f_timer)
        self.timer.addSubTimer(self.kernels_timer, 'Details:')
        ParticleAdvection.finalize(self)
