"""
@file parmepy.gpu.tests.test_advection_nullVelocity
Testing advection kernels with a null velocity. Basic functionnal test.
"""
from parmepy.domain.box import Box
from parmepy.fields.continuous import Field
from parmepy.operator.advection import Advection
from parmepy.constants import ORDER, np, PARMES_REAL
from parmepy.problem.simulation import Simulation
from parmepy.methods_keys import TimeIntegrator, Interpolation, Remesh, \
    Support, Splitting, Precision
from parmepy.numerics.integrators.runge_kutta2 import RK2
from parmepy.numerics.interpolation import Linear
from parmepy.numerics.remeshing import L2_1, L4_2, L6_3, M8Prime


def setup_2D():
    box = Box(2, length=[1., 1.], origin=[0., 0.])
    scal = Field(domain=box, name='Scalar')
    velo = Field(domain=box, name='Velocity', isVector=True)
    return scal, velo


def setup_3D():
    box = Box(3, length=[1., 1., 1.], origin=[0., 0., 0.])
    scal = Field(domain=box, name='Scalar')
    velo = Field(domain=box, name='Velocity', isVector=True)
    return scal, velo


def assertion_2D(scal, velo, advec):
    advec.discretize()
    advec.setUp()

    scal_d = scal.discreteFields.values()[0]
    velo_d = velo.discreteFields.values()[0]
    scal_d.data[0][...] = np.asarray(np.random.random(scal_d.data[0].shape),
                                     dtype=PARMES_REAL, order=ORDER)
    velo_d.data[0][...] = np.zeros_like(scal_d.data[0],
                                        dtype=PARMES_REAL, order=ORDER)
    velo_d.data[1][...] = np.zeros_like(scal_d.data[0],
                                        dtype=PARMES_REAL, order=ORDER)
    scal_init = scal_d.data[0].copy()
    scal_d.toDevice()
    velo_d.toDevice()

    advec.apply(Simulation(tinit=0., tend=0.01, nbIter=1))

    scal_d.toHost()

    advec.finalize()
    return np.allclose(scal_init, scal_d.data[0])


def assertion_2D_withPython(scal, velo, advec, advec_py):
    advec.discretize()
    advec_py.discretize()
    advec.setUp()
    advec_py.setUp()

    scal_d = scal.discreteFields.values()[0]
    velo_d = velo.discreteFields.values()[0]
    scal_d.data[0][...] = np.asarray(np.random.random(scal_d.data[0].shape),
                                     dtype=PARMES_REAL, order=ORDER)
    velo_d.data[0][...] = np.zeros_like(scal_d.data[0],
                                        dtype=PARMES_REAL, order=ORDER)
    velo_d.data[1][...] = np.zeros_like(scal_d.data[0],
                                        dtype=PARMES_REAL, order=ORDER)
    scal_d.toDevice()
    velo_d.toDevice()

    advec_py.apply(Simulation(tinit=0., tend=0.01, nbIter=1))
    advec.apply(Simulation(tinit=0., tend=0.01, nbIter=1))

    py_res = scal_d.data[0].copy()
    scal_d.toHost()

    advec.finalize()
    print py_res, scal_d.data[0]
    print py_res - scal_d.data[0]
    return np.allclose(py_res, scal_d.data[0])


def assertion_3D(scal, velo, advec):
    advec.discretize()
    advec.setUp()

    scal_d = scal.discreteFields.values()[0]
    velo_d = velo.discreteFields.values()[0]
    scal_d.data[0][...] = np.asarray(np.random.random(scal_d.data[0].shape),
                                     dtype=PARMES_REAL, order=ORDER)
    velo_d.data[0][...] = np.zeros_like(scal_d.data[0],
                                        dtype=PARMES_REAL, order=ORDER)
    velo_d.data[1][...] = np.zeros_like(scal_d.data[0],
                                        dtype=PARMES_REAL, order=ORDER)
    velo_d.data[2][...] = np.zeros_like(scal_d.data[0],
                                        dtype=PARMES_REAL, order=ORDER)
    scal_init = scal_d.data[0].copy()
    scal_d.toDevice()
    velo_d.toDevice()

    advec.apply(Simulation(tinit=0., tend=0.01, nbIter=1))

    scal_d.toHost()

    advec.finalize()
    return np.allclose(scal_init, scal_d.data[0])


def assertion_3D_withPython(scal, velo, advec, advec_py):
    advec.discretize()
    advec_py.discretize()
    advec.setUp()
    advec_py.setUp()

    scal_d = scal.discreteFields.values()[0]
    velo_d = velo.discreteFields.values()[0]
    scal_d.data[0][...] = np.asarray(np.random.random(scal_d.data[0].shape),
                                     dtype=PARMES_REAL, order=ORDER)
    velo_d.data[0][...] = np.zeros_like(scal_d.data[0],
                                        dtype=PARMES_REAL, order=ORDER)
    velo_d.data[1][...] = np.zeros_like(scal_d.data[0],
                                        dtype=PARMES_REAL, order=ORDER)
    velo_d.data[2][...] = np.zeros_like(scal_d.data[0],
                                        dtype=PARMES_REAL, order=ORDER)
    scal_d.toDevice()
    velo_d.toDevice()

    advec_py.apply(Simulation(tinit=0., tend=0.01, nbIter=1))
    advec.apply(Simulation(tinit=0., tend=0.01, nbIter=1))

    py_res = scal_d.data[0].copy()
    scal_d.toHost()

    advec.finalize()
    return np.allclose(py_res, scal_d.data[0])


# M6 tests
def test_2D_m6_1k():
    """
    Testing M6 remeshing formula in 2D, 1 kernel, simple precision,
    o2 splitting.
    """
    scal, velo = setup_2D()

    advec = Advection(velo, scal,
                      resolutions={velo: [33, 33],
                                   scal: [33, 33]},
                      method={TimeIntegrator: RK2,
                              Interpolation: Linear,
                              Remesh: L4_2,
                              Support: 'gpu_1k',
                              Splitting: 'o2',
                              Precision: PARMES_REAL}
                      )
    advec_py = Advection(velo, scal,
                         resolutions={velo: [33, 33],
                                      scal: [33, 33]},
                         method={TimeIntegrator: RK2,
                                 Interpolation: Linear,
                                 Remesh: L4_2,
                                 Support: '',
                                 Splitting: 'o2'},
                         )
    assert assertion_2D_withPython(scal, velo, advec, advec_py)
    #assert assertion_2D(scal, velo, advec)


def test_2D_m6_2k():
    """
    Testing M6 remeshing formula in 2D, 2 kernel, simple precision,
    o2 splitting.
    """
    scal, velo = setup_2D()

    advec = Advection(velo, scal,
                      resolutions={velo: [33, 33],
                                   scal: [33, 33]},
                      method={TimeIntegrator: RK2,
                              Interpolation: Linear,
                              Remesh: L4_2,
                              Support: 'gpu_2k',
                              Splitting: 'o2',
                              Precision: PARMES_REAL}
                      )
    assert assertion_2D(scal, velo, advec)


def test_2D_m6_1k_sFH():
    """
    Testing M6 remeshing formula in 2D, 1 kernel, simple precision,
    o2_FullHalf splitting.
    """
    scal, velo = setup_2D()

    advec = Advection(velo, scal,
                      resolutions={velo: [33, 33],
                                   scal: [33, 33]},
                      method={TimeIntegrator: RK2,
                              Interpolation: Linear,
                              Remesh: L4_2,
                              Support: 'gpu_1k',
                              Splitting: 'o2_FullHalf',
                              Precision: PARMES_REAL}
                      )
    assert assertion_2D(scal, velo, advec)


def test_2D_m6_2k_sFH():
    """
    Testing M6 remeshing formula in 2D, 2 kernel, simple precision,
    o2_FullHalf splitting.
    """
    scal, velo = setup_2D()

    advec = Advection(velo, scal,
                      resolutions={velo: [33, 33],
                                   scal: [33, 33]},
                      method={TimeIntegrator: RK2,
                              Interpolation: Linear,
                              Remesh: L4_2,
                              Support: 'gpu_2k',
                              Splitting: 'o2_FullHalf',
                              Precision: PARMES_REAL}
                      )
    assert assertion_2D(scal, velo, advec)


def test_3D_m6_1k():
    """
    Testing M6 remeshing formula in 3D, 1 kernel, simple precision,
    o2 splitting.
    """
    scal, velo = setup_3D()

    advec = Advection(velo, scal,
                      resolutions={velo: [17, 17, 17],
                                   scal: [17, 17, 17]},
                      method={TimeIntegrator: RK2,
                              Interpolation: Linear,
                              Remesh: L4_2,
                              Support: 'gpu_1k',
                              Splitting: 'o2',
                              Precision: PARMES_REAL}
                      )
    assert assertion_3D(scal, velo, advec)


def test_3D_m6_2k():
    """
    Testing M6 remeshing formula in 3D, 2 kernel, simple precision,
    o2 splitting.
    """
    scal, velo = setup_3D()

    advec = Advection(velo, scal,
                      resolutions={velo: [17, 17, 17],
                                   scal: [17, 17, 17]},
                      method={TimeIntegrator: RK2,
                              Interpolation: Linear,
                              Remesh: L4_2,
                              Support: 'gpu_2k',
                              Splitting: 'o2',
                              Precision: PARMES_REAL}
                        )
    assert assertion_3D(scal, velo, advec)


def test_3D_m6_1k_sFH():
    """
    Testing M6 remeshing formula in 3D, 1 kernel, simple precision,
    o2_FullHalf splitting.
    """
    scal, velo = setup_3D()

    advec = Advection(velo, scal,
                      resolutions={velo: [17, 17, 17],
                                   scal: [17, 17, 17]},
                      method={TimeIntegrator: RK2,
                              Interpolation: Linear,
                              Remesh: L4_2,
                              Support: 'gpu_1k',
                              Splitting: 'o2_FullHalf',
                              Precision: PARMES_REAL}
                      )
    assert assertion_3D(scal, velo, advec)


def test_3D_m6_2k_sFH():
    """
    Testing M6 remeshing formula in 3D, 2 kernel, simple precision,
    o2_FullHalf splitting.
    """
    scal, velo = setup_3D()

    advec = Advection(velo, scal,
                      resolutions={velo: [17, 17, 17],
                                   scal: [17, 17, 17]},
                      method={TimeIntegrator: RK2,
                              Interpolation: Linear,
                              Remesh: L4_2,
                              Support: 'gpu_2k',
                              Splitting: 'o2_FullHalf',
                              Precision: PARMES_REAL},
                      )
    assert assertion_3D(scal, velo, advec)


# M4 testing
def test_2D_m4_1k():
    """
    Testing M4 remeshing formula in 2D, 1 kernel, simple precision,
    o2 splitting.
    """
    scal, velo = setup_2D()

    advec = Advection(velo, scal,
                      resolutions={velo: [33, 33],
                                   scal: [33, 33]},
                      method={TimeIntegrator: RK2,
                              Interpolation: Linear,
                              Remesh: L2_1,
                              Support: 'gpu_1k',
                              Splitting: 'o2',
                              Precision: PARMES_REAL}
                        )
    advec_py = Advection(velo, scal,
                         resolutions={velo: [33, 33],
                                      scal: [33, 33]},
                         method={TimeIntegrator: RK2,
                                 Interpolation: Linear,
                                 Remesh: L2_1,
                                 Support: '',
                                 Splitting: 'o2'}
                         )
    assert assertion_2D_withPython(scal, velo, advec, advec_py)


def test_2D_m4_2k():
    """
    Testing M4 remeshing formula in 2D, 2 kernel, simple precision,
    o2 splitting.
    """
    scal, velo = setup_2D()

    advec = Advection(velo, scal,
                      resolutions={velo: [33, 33],
                                   scal: [33, 33]},
                      method={TimeIntegrator: RK2,
                              Interpolation: Linear,
                              Remesh: L2_1,
                              Support: 'gpu_2k',
                              Splitting: 'o2',
                              Precision: PARMES_REAL
                              }
                        )
    advec_py = Advection(velo, scal,
                         resolutions={velo: [33, 33],
                                      scal: [33, 33]},
                         method={TimeIntegrator: RK2,
                                 Interpolation: Linear,
                                 Remesh: L2_1,
                                 Support: '',
                                 Splitting: 'o2'},
                         )
    assert assertion_2D_withPython(scal, velo, advec, advec_py)


def test_2D_m4_1k_sFH():
    """
    Testing M4 remeshing formula in 2D, 1 kernel, simple precision,
    o2_FullHalf splitting.
    """
    scal, velo = setup_2D()

    advec = Advection(velo, scal,
                      resolutions={velo: [33, 33],
                                   scal: [33, 33]},
                      method={TimeIntegrator: RK2,
                              Interpolation: Linear,
                              Remesh: L2_1,
                              Support: 'gpu_1k',
                              Splitting: 'o2_FullHalf',
                              Precision: PARMES_REAL},
                        )
    advec_py = Advection(velo, scal,
                         resolutions={velo: [33, 33],
                                      scal: [33, 33]},
                         method={TimeIntegrator: RK2,
                                 Interpolation: Linear,
                                 Remesh: L2_1,
                                 Support: '',
                                 Splitting: 'o2_FullHalf'}
                         )
    assert assertion_2D_withPython(scal, velo, advec, advec_py)


def test_2D_m4_2k_sFH():
    """
    Testing M4 remeshing formula in 2D, 2 kernel, simple precision,
    o2_FullHalf splitting.
    """
    scal, velo = setup_2D()

    advec = Advection(velo, scal,
                      resolutions={velo: [33, 33],
                                   scal: [33, 33]},
                      method={TimeIntegrator: RK2,
                              Interpolation: Linear,
                              Remesh: L2_1,
                              Support: 'gpu_2k',
                              Splitting: 'o2_FullHalf'}
                      )
    advec_py = Advection(velo, scal,
                         resolutions={velo: [33, 33],
                                      scal: [33, 33]},
                         method={TimeIntegrator: RK2,
                                 Interpolation: Linear,
                                 Remesh: L2_1,
                                 Support: '',
                                 Splitting: 'o2_FullHalf'}
                         )
    assert assertion_2D_withPython(scal, velo, advec, advec_py)


def test_3D_m4_1k():
    """
    Testing M4 remeshing formula in 3D, 1 kernel, simple precision,
    o2 splitting.
    """
    scal, velo = setup_3D()

    advec = Advection(velo, scal,
                      resolutions={velo: [17, 17, 17],
                                   scal: [17, 17, 17]},
                      method={TimeIntegrator: RK2,
                              Interpolation: Linear,
                              Remesh: L2_1,
                              Support: 'gpu_1k',
                              Splitting: 'o2'}
                        )
    advec_py = Advection(velo, scal,
                         resolutions={velo: [17, 17, 17],
                                      scal: [17, 17, 17]},
                         method={TimeIntegrator: RK2,
                                 Interpolation: Linear,
                                 Remesh: L2_1,
                                 Support: '',
                                 Splitting: 'o2'},
                         )
    assert assertion_3D_withPython(scal, velo, advec, advec_py)


def test_3D_m4_2k():
    """
    Testing M4 remeshing formula in 3D, 2 kernel, simple precision,
    o2 splitting.
    """
    scal, velo = setup_3D()

    advec = Advection(velo, scal,
                      resolutions={velo: [17, 17, 17],
                                   scal: [17, 17, 17]},
                      method={TimeIntegrator: RK2,
                              Interpolation: Linear,
                              Remesh: L2_1,
                              Support: 'gpu_2k',
                              Splitting: 'o2'}
                            )
    advec_py = Advection(velo, scal,
                         resolutions={velo: [17, 17, 17],
                                      scal: [17, 17, 17]},
                         method={TimeIntegrator: RK2,
                                 Interpolation: Linear,
                                 Remesh: L2_1,
                                 Support: '',
                                 Splitting: 'o2'},
                         )
    assert assertion_3D_withPython(scal, velo, advec, advec_py)


def test_3D_m4_1k_sFH():
    """
    Testing M4 remeshing formula in 3D, 1 kernel, simple precision,
    o2_FullHalf splitting.
    """
    scal, velo = setup_3D()

    advec = Advection(velo, scal,
                      resolutions={velo: [17, 17, 17],
                                   scal: [17, 17, 17]},
                      method={TimeIntegrator: RK2,
                              Interpolation: Linear,
                              Remesh: L2_1,
                              Support: 'gpu_1k',
                              Splitting: 'o2_FullHalf'})
    advec_py = Advection(velo, scal,
                         resolutions={velo: [17, 17, 17],
                                      scal: [17, 17, 17]},
                         method={TimeIntegrator: RK2,
                                 Interpolation: Linear,
                                 Remesh: L2_1,
                                 Support: '',
                                 Splitting: 'o2_FullHalf'}
                         )
    assert assertion_3D_withPython(scal, velo, advec, advec_py)


def test_3D_m4_2k_sFH():
    """
    Testing M4 remeshing formula in 3D, 2 kernel, simple precision,
    o2_FullHalf splitting.
    """
    scal, velo = setup_3D()

    advec = Advection(velo, scal,
                      resolutions={velo: [17, 17, 17],
                                   scal: [17, 17, 17]},
                      method={TimeIntegrator: RK2,
                              Interpolation: Linear,
                              Remesh: L2_1,
                              Support: 'gpu_2k',
                              Splitting: 'o2_FullHalf'}
                      )
    advec_py = Advection(velo, scal,
                         resolutions={velo: [17, 17, 17],
                                      scal: [17, 17, 17]},
                         method={TimeIntegrator: RK2,
                                 Interpolation: Linear,
                                 Remesh: L2_1,
                                 Support: '',
                                 Splitting: 'o2_FullHalf'}
                         )
    assert assertion_3D_withPython(scal, velo, advec, advec_py)


# M8 testing
def test_2D_m8_1k():
    """
    Testing M8 remeshing formula in 2D, 1 kernel, simple precision,
    o2 splitting.
    """
    scal, velo = setup_2D()

    advec = Advection(velo, scal,
                      resolutions={velo: [33, 33],
                                   scal: [33, 33]},
                      method={TimeIntegrator: RK2,
                              Interpolation: Linear,
                              Remesh: M8Prime,
                              Support: 'gpu_1k',
                              Splitting: 'o2'}
                      )
    advec_py = Advection(velo, scal,
                         resolutions={velo: [33, 33],
                                      scal: [33, 33]},
                         method={TimeIntegrator: RK2,
                                 Interpolation: Linear,
                                 Remesh: M8Prime,
                                 Support: '',
                                 Splitting: 'o2'},
                         )
    assert assertion_2D_withPython(scal, velo, advec, advec_py)


def test_2D_m8_2k():
    """
    Testing M8 remeshing formula in 2D, 2 kernel, simple precision,
    o2 splitting.
    """
    scal, velo = setup_2D()

    advec = Advection(velo, scal,
                      resolutions={velo: [33, 33],
                                   scal: [33, 33]},
                      method={TimeIntegrator: RK2,
                              Interpolation: Linear,
                              Remesh: M8Prime,
                              Support: 'gpu_2k',
                              Splitting: 'o2'}
                      )
    advec_py = Advection(velo, scal,
                         resolutions={velo: [33, 33],
                                      scal: [33, 33]},
                         method={TimeIntegrator: RK2,
                                 Interpolation: Linear,
                                 Remesh: M8Prime,
                                 Support: '',
                                 Splitting: 'o2'},
                         )
    assert assertion_2D_withPython(scal, velo, advec, advec_py)


def test_2D_m8_1k_sFH():
    """
    Testing M8 remeshing formula in 2D, 1 kernel, simple precision,
    o2_FullHalf splitting.
    """
    scal, velo = setup_2D()

    advec = Advection(velo, scal,
                      resolutions={velo: [33, 33],
                                   scal: [33, 33]},
                      method={TimeIntegrator: RK2,
                              Interpolation: Linear,
                              Remesh: M8Prime,
                              Support: 'gpu_1k',
                              Splitting: 'o2_FullHalf'}
                      )
    advec_py = Advection(velo, scal,
                         resolutions={velo: [33, 33],
                                      scal: [33, 33]},
                         method={TimeIntegrator: RK2,
                                 Interpolation: Linear,
                                 Remesh: M8Prime,
                                 Support: '',
                                 Splitting: 'o2_FullHalf'}
                         )
    assert assertion_2D_withPython(scal, velo, advec, advec_py)


def test_2D_m8_2k_sFH():
    """
    Testing M8 remeshing formula in 2D, 2 kernel, simple precision,
    o2_FullHalf splitting.
    """
    scal, velo = setup_2D()

    advec = Advection(velo, scal,
                      resolutions={velo: [33, 33],
                                   scal: [33, 33]},
                      method={TimeIntegrator: RK2,
                              Interpolation: Linear,
                              Remesh: M8Prime,
                              Support: 'gpu_2k',
                              Splitting: 'o2_FullHalf'}
                      )
    advec_py = Advection(velo, scal,
                         resolutions={velo: [33, 33],
                                      scal: [33, 33]},
                         method={TimeIntegrator: RK2,
                                 Interpolation: Linear,
                                 Remesh: M8Prime,
                                 Support: '',
                                 Splitting: 'o2_FullHalf'}
                         )
    assert assertion_2D_withPython(scal, velo, advec, advec_py)


def test_3D_m8_1k():
    """
    Testing M8 remeshing formula in 3D, 1 kernel, simple precision,
    o2 splitting.
    """
    scal, velo = setup_3D()

    advec = Advection(velo, scal,
                      resolutions={velo: [17, 17, 17],
                                   scal: [17, 17, 17]},
                      method={TimeIntegrator: RK2,
                              Interpolation: Linear,
                              Remesh: M8Prime,
                              Support: 'gpu_1k',
                              Splitting: 'o2'}
                      )
    advec_py = Advection(velo, scal,
                         resolutions={velo: [17, 17, 17],
                                      scal: [17, 17, 17]},
                         method={TimeIntegrator: RK2,
                                 Interpolation: Linear,
                                 Remesh: M8Prime,
                                 Support: '',
                                 Splitting: 'o2'},
                         )
    assert assertion_3D_withPython(scal, velo, advec, advec_py)


def test_3D_m8_2k():
    """
    Testing M8 remeshing formula in 3D, 2 kernel, simple precision,
    o2 splitting.
    """
    scal, velo = setup_3D()

    advec = Advection(velo, scal,
                      resolutions={velo: [17, 17, 17],
                                   scal: [17, 17, 17]},
                      method={TimeIntegrator: RK2,
                              Interpolation: Linear,
                              Remesh: M8Prime,
                              Support: 'gpu_2k',
                              Splitting: 'o2'}
                      )
    advec_py = Advection(velo, scal,
                         resolutions={velo: [17, 17, 17],
                                      scal: [17, 17, 17]},
                         method={TimeIntegrator: RK2,
                                 Interpolation: Linear,
                                 Remesh: M8Prime,
                                 Support: '',
                                 Splitting: 'o2'},
                         )
    assert assertion_3D_withPython(scal, velo, advec, advec_py)


def test_3D_m8_1k_sFH():
    """
    Testing M8 remeshing formula in 3D, 1 kernel, simple precision,
    o2_FullHalf splitting.
    """
    scal, velo = setup_3D()

    advec = Advection(velo, scal,
                      resolutions={velo: [17, 17, 17],
                                   scal: [17, 17, 17]},
                      method={TimeIntegrator: RK2,
                              Interpolation: Linear,
                              Remesh: M8Prime,
                              Support: 'gpu_1k',
                              Splitting: 'o2_FullHalf'}
                      )
    advec_py = Advection(velo, scal,
                         resolutions={velo: [17, 17, 17],
                                      scal: [17, 17, 17]},
                         method={TimeIntegrator: RK2,
                                 Interpolation: Linear,
                                 Remesh: M8Prime,
                                 Support: '',
                                 Splitting: 'o2_FullHalf'}
                         )
    assert assertion_3D_withPython(scal, velo, advec, advec_py)


def test_3D_m8_2k_sFH():
    """
    Testing M8 remeshing formula in 3D, 2 kernel, simple precision,
    o2_FullHalf splitting.
    """
    scal, velo = setup_3D()

    advec = Advection(velo, scal,
                      resolutions={velo: [17, 17, 17],
                                   scal: [17, 17, 17]},
                      method={TimeIntegrator: RK2,
                              Interpolation: Linear,
                              Remesh: M8Prime,
                              Support: 'gpu_2k',
                              Splitting: 'o2_FullHalf'}
                      )
    advec_py = Advection(velo, scal,
                         resolutions={velo: [17, 17, 17],
                                      scal: [17, 17, 17]},
                         method={TimeIntegrator: RK2,
                                 Interpolation: Linear,
                                 Remesh: M8Prime,
                                 Support: '',
                                 Splitting: 'o2_FullHalf'}
                         )
    assert assertion_3D_withPython(scal, velo, advec, advec_py)


def test_2D_l6_2k():
    """
    Testing Lambda6star remeshing formula in 2D, 2 kernel, simple precision,
    o2 splitting.
    """
    scal, velo = setup_2D()

    advec = Advection(velo, scal,
                      resolutions={velo: [33, 33],
                                   scal: [33, 33]},
                      method={TimeIntegrator: RK2,
                              Interpolation: Linear,
                              Remesh: L6_3,
                              Support: 'gpu_1k',
                              Splitting: 'o2'}
                      )
    assert assertion_2D(scal, velo, advec)


def test_2D_l6_1k_sFH():
    """
    Testing Lambda6star remeshing formula in 2D, 1 kernel, simple precision,
    o2_FullHalf splitting.
    """
    scal, velo = setup_2D()

    advec = Advection(velo, scal,
                      resolutions={velo: [33, 33],
                                   scal: [33, 33]},
                      method={TimeIntegrator: RK2,
                              Interpolation: Linear,
                              Remesh: L6_3,
                              Support: 'gpu_1k',
                              Splitting: 'o2_FullHalf'}
                      )
    assert assertion_2D(scal, velo, advec)


def test_2D_l6_2k_sFH():
    """
    Testing Lambda6star remeshing formula in 2D, 2 kernel, simple precision,
    o2_FullHalf splitting.
    """
    scal, velo = setup_2D()

    advec = Advection(velo, scal,
                      resolutions={velo: [33, 33],
                                   scal: [33, 33]},
                      method={TimeIntegrator: RK2,
                              Interpolation: Linear,
                              Remesh: L6_3,
                              Support: 'gpu_2k',
                              Splitting: 'o2_FullHalf'}
                      )
    assert assertion_2D(scal, velo, advec)


def test_3D_l6_1k():
    """
    Testing Lambda6star remeshing formula in 3D, 1 kernel, simple precision,
    o2 splitting.
    """
    scal, velo = setup_3D()

    advec = Advection(velo, scal,
                      resolutions={velo: [17, 17, 17],
                                   scal: [17, 17, 17]},
                      method={TimeIntegrator: RK2,
                              Interpolation: Linear,
                              Remesh: L6_3,
                              Support: 'gpu_1k',
                              Splitting: 'o2'}
                      )
    assert assertion_3D(scal, velo, advec)


def test_3D_l6_2k():
    """
    Testing Lambda6star remeshing formula in 3D, 2 kernel, simple precision,
    o2 splitting.
    """
    scal, velo = setup_3D()

    advec = Advection(velo, scal,
                      resolutions={velo: [17, 17, 17],
                                   scal: [17, 17, 17]},
                      method={TimeIntegrator: RK2,
                              Interpolation: Linear,
                              Remesh: L6_3,
                              Support: 'gpu_2k',
                              Splitting: 'o2'}
                      )
    assert assertion_3D(scal, velo, advec)


def test_3D_l6_1k_sFH():
    """
    Testing Lambda6star remeshing formula in 3D, 1 kernel, simple precision,
    o2_FullHalf splitting.
    """
    scal, velo = setup_3D()

    advec = Advection(velo, scal,
                      resolutions={velo: [17, 17, 17],
                                   scal: [17, 17, 17]},
                      method={TimeIntegrator: RK2,
                              Interpolation: Linear,
                              Remesh: L6_3,
                              Support: 'gpu_1k',
                              Splitting: 'o2_FullHalf'}
                      )
    assert assertion_3D(scal, velo, advec)


def test_3D_l6_2k_sFH():
    """
    Testing Lambda6star remeshing formula in 3D, 2 kernel, simple precision,
    o2_FullHalf splitting.
    """
    scal, velo = setup_3D()

    advec = Advection(velo, scal,
                      resolutions={velo: [17, 17, 17],
                                   scal: [17, 17, 17]},
                      method={TimeIntegrator: RK2,
                              Interpolation: Linear,
                              Remesh: L6_3,
                              Support: 'gpu_2k',
                              Splitting: 'o2_FullHalf'}
                      )
    assert assertion_3D(scal, velo, advec)


def test_rectangular_domain2D():
    box = Box(2, length=[1., 1.], origin=[0., 0.])
    scal = Field(domain=box, name='Scalar')
    velo = Field(domain=box, name='Velocity', isVector=True)

    advec = Advection(velo, scal,
                      resolutions={velo: [65, 33],
                                   scal: [65, 33]},
                      method={TimeIntegrator: RK2,
                              Interpolation: Linear,
                              Remesh: L4_2,
                              Support: 'gpu_1k',
                              Splitting: 'o2'}
                      )
    advec_py = Advection(velo, scal,
                         resolutions={velo: [65, 33],
                                      scal: [65, 33]},
                         method={TimeIntegrator: RK2,
                                 Interpolation: Linear,
                                 Remesh: L4_2,
                                 Support: '',
                                 Splitting: 'o2'},
                         )
    advec.discretize()
    advec_py.discretize()
    advec.setUp()
    advec_py.setUp()

    scal_d = scal.discreteFields.values()[0]
    velo_d = velo.discreteFields.values()[0]
    scal_d.data[0][...] = np.asarray(np.random.random(scal_d.data[0].shape),
                                     dtype=PARMES_REAL, order=ORDER)
    velo_d.data[0][...] = np.zeros_like(scal_d.data[0],
                                        dtype=PARMES_REAL, order=ORDER)
    velo_d.data[1][...] = np.zeros_like(scal_d.data[0],
                                        dtype=PARMES_REAL, order=ORDER)
    scal_init = scal_d.data[0].copy()
    scal_d.toDevice()
    velo_d.toDevice()

    advec.apply(Simulation(tinit=0., tend=0.01, nbIter=1))
    advec_py.apply(Simulation(tinit=0., tend=0.01, nbIter=1))
    scal_py_res = scal_d.data[0].copy()

    scal_d.toHost()

    advec.finalize()
    assert np.allclose(scal_init, scal_d.data[0])
    assert np.allclose(scal_init, scal_py_res)


def test_rectangular_domain3D():
    box = Box(3, length=[1., 1., 1.], origin=[0., 0., 0.])
    scal = Field(domain=box, name='Scalar')
    velo = Field(domain=box, name='Velocity', isVector=True)

    advec = Advection(velo, scal,
                      resolutions={velo: [65, 33, 17],
                                   scal: [65, 33, 17]},
                      method={TimeIntegrator: RK2,
                              Interpolation: Linear,
                              Remesh: L4_2,
                              Support: 'gpu_1k',
                              Splitting: 'o2'}
                      )
    advec_py = Advection(velo, scal,
                         resolutions={velo: [65, 33, 17],
                                      scal: [65, 33, 17]},
                         method={TimeIntegrator: RK2,
                                 Interpolation: Linear,
                                 Remesh: L4_2,
                                 Support: '',
                                 Splitting: 'o2'},
                         )
    advec.discretize()
    advec_py.discretize()
    advec.setUp()
    advec_py.setUp()

    scal_d = scal.discreteFields.values()[0]
    velo_d = velo.discreteFields.values()[0]
    scal_d.data[0][...] = np.asarray(np.random.random(scal_d.data[0].shape),
                                     dtype=PARMES_REAL, order=ORDER)
    velo_d.data[0][...] = np.zeros_like(scal_d.data[0],
                                        dtype=PARMES_REAL, order=ORDER)
    velo_d.data[1][...] = np.zeros_like(scal_d.data[0],
                                        dtype=PARMES_REAL, order=ORDER)
    velo_d.data[2][...] = np.zeros_like(scal_d.data[0],
                                        dtype=PARMES_REAL, order=ORDER)
    scal_init = scal_d.data[0].copy()
    scal_d.toDevice()
    velo_d.toDevice()

    advec.apply(Simulation(tinit=0., tend=0.01, nbIter=1))
    advec_py.apply(Simulation(tinit=0., tend=0.01, nbIter=1))
    scal_py_res = scal_d.data[0].copy()

    scal_d.toHost()

    advec.finalize()
    assert np.allclose(scal_init, scal_d.data[0])
    assert np.allclose(scal_init, scal_py_res)


def test_2D_vector():
    box = Box(2, length=[1., 1.], origin=[0., 0.])
    scal = Field(domain=box, name='Scalar', isVector=True)
    velo = Field(domain=box, name='Velocity', isVector=True)

    advec = Advection(velo, scal,
                      resolutions={velo: [129, 129],
                                   scal: [129, 129]},
                      method={TimeIntegrator: RK2,
                              Interpolation: Linear,
                              Remesh: L4_2,
                              Support: 'gpu_1k',
                              Splitting: 'o2'}
                      )
    advec_py = Advection(velo, scal,
                         resolutions={velo: [129, 129],
                                      scal: [129, 129]},
                         method={TimeIntegrator: RK2,
                                 Interpolation: Linear,
                                 Remesh: L4_2,
                                 Support: '',
                                 Splitting: 'o2'},
                         )
    advec.discretize()
    advec_py.discretize()
    advec.setUp()
    advec_py.setUp()

    scal_d = scal.discreteFields.values()[0]
    velo_d = velo.discreteFields.values()[0]
    scal_d.data[0][...] = np.asarray(np.random.random(scal_d.data[0].shape),
                                     dtype=PARMES_REAL, order=ORDER)
    scal_d.data[1][...] = np.asarray(np.random.random(scal_d.data[0].shape),
                                     dtype=PARMES_REAL, order=ORDER)
    velo_d.data[0][...] = np.zeros_like(scal_d.data[0],
                                        dtype=PARMES_REAL, order=ORDER)
    velo_d.data[1][...] = np.zeros_like(scal_d.data[0],
                                        dtype=PARMES_REAL, order=ORDER)
    scal_init_X = scal_d.data[0].copy()
    scal_init_Y = scal_d.data[1].copy()
    scal_d.toDevice()
    velo_d.toDevice()

    advec.apply(Simulation(tinit=0., tend=0.01, nbIter=1))
    advec_py.apply(Simulation(tinit=0., tend=0.01, nbIter=1))
    scal_py_res_X = scal_d.data[0].copy()
    scal_py_res_Y = scal_d.data[1].copy()

    scal_d.toHost()

    advec.finalize()
    assert np.allclose(scal_init_X, scal_d.data[0])
    assert np.allclose(scal_init_Y, scal_d.data[1])
    assert np.allclose(scal_init_X, scal_py_res_X)
    assert np.allclose(scal_init_Y, scal_py_res_Y)


def test_3D_vector():
    box = Box(3, length=[1., 1., 1.], origin=[0., 0., 0.])
    scal = Field(domain=box, name='Scalar', isVector=True)
    velo = Field(domain=box, name='Velocity', isVector=True)

    advec = Advection(velo, scal,
                      resolutions={velo: [17, 17, 17],
                                   scal: [17, 17, 17]},
                      method={TimeIntegrator: RK2,
                              Interpolation: Linear,
                              Remesh: L4_2,
                              Support: 'gpu_1k',
                              Splitting: 'o2'}
                        )
    advec_py = Advection(velo, scal,
                         resolutions={velo: [17, 17, 17],
                                      scal: [17, 17, 17]},
                         method={TimeIntegrator: RK2,
                                 Interpolation: Linear,
                                 Remesh: L4_2,
                                 Support: '',
                                 Splitting: 'o2'},
                         )
    advec.discretize()
    advec_py.discretize()
    advec.setUp()
    advec_py.setUp()

    scal_d = scal.discreteFields.values()[0]
    velo_d = velo.discreteFields.values()[0]
    scal_d.data[0][...] = np.asarray(np.random.random(scal_d.data[0].shape),
                                     dtype=PARMES_REAL, order=ORDER)
    scal_d.data[1][...] = np.asarray(np.random.random(scal_d.data[0].shape),
                                     dtype=PARMES_REAL, order=ORDER)
    scal_d.data[2][...] = np.asarray(np.random.random(scal_d.data[0].shape),
                                     dtype=PARMES_REAL, order=ORDER)
    velo_d.data[0][...] = np.zeros_like(scal_d.data[0],
                                        dtype=PARMES_REAL, order=ORDER)
    velo_d.data[1][...] = np.zeros_like(scal_d.data[0],
                                        dtype=PARMES_REAL, order=ORDER)
    velo_d.data[2][...] = np.zeros_like(scal_d.data[0],
                                        dtype=PARMES_REAL, order=ORDER)
    scal_init_X = scal_d.data[0].copy()
    scal_init_Y = scal_d.data[1].copy()
    scal_init_Z = scal_d.data[2].copy()
    scal_d.toDevice()
    velo_d.toDevice()

    advec.apply(Simulation(tinit=0., tend=0.01, nbIter=1))
    advec_py.apply(Simulation(tinit=0., tend=0.01, nbIter=1))
    scal_py_res_X = scal_d.data[0].copy()
    scal_py_res_Y = scal_d.data[1].copy()
    scal_py_res_Z = scal_d.data[2].copy()

    scal_d.toHost()

    advec.finalize()
    assert np.allclose(scal_init_X, scal_d.data[0])
    assert np.allclose(scal_init_Y, scal_d.data[1])
    assert np.allclose(scal_init_Z, scal_d.data[2])
    assert np.allclose(scal_init_X, scal_py_res_X)
    assert np.allclose(scal_init_Y, scal_py_res_Y)
    assert np.allclose(scal_init_Z, scal_py_res_Z)
