"""
@file bridge.py
Tools to compute the intersection between
two topologies.
"""
from parmepy.constants import np, debug
from parmepy.mpi.main_var import MPI
import parmepy.tools.numpywrappers as npw
from parmepy.mpi.topology import topotools


class Bridge(object):
    """
    A Bridge is the definition of what must be exchanged between two topologies
    and how to do so.

    For a bridge between topo1 and topo2, it will provide :
    bridge.sendTo = [[rk, i, j, k, l, ...]]
    bridge.recvFrom = [[rk, i, j, k, l, ...]]
    where each line corresponds to a message that must be send
    from current mpi process to process rk for sendTo or
    receive by current mpi process from process rk for recvFrom.
    The other columns give the numbers of the node of the mesh
    that must be exchanged.
    Warning : index numbers corresponds to local nodes numbers.

    In sendTo/recvFrom :
    * First column : dest/source of the message.
    Warning : rank number is given in main_comm.
    * second-third columns : first and last index of nodes in the first dir
    * 4th-5th : first and last index of nodes in the second dir
    * and so on depending on the size of the domain
    associated to the topologies.

    """

    @debug
    def __init__(self, topoFrom, topoTo, parentComm=None):
        """
        Bridge constructor.
        @param source topology
        @param targeted topology
        @param parent communicator, if None, use topoFrom.parent()
        """
        ## Source topology
        self.topoFrom = topoFrom
        ## Targeted topology
        self.topoTo = topoTo

        # Both topologies must be uptodate
        assert(topoFrom.isUpToDate)
        assert(topoTo.isUpToDate)

        ## recvFrom[rk] returns a list of indices that should be
        ## received from process rk and copied into local field.
        ## Usage :  mpi.recv(rk, ..., yourField[recvFrom[rk]],...)
        self.recvFrom = {}

        ## sendTo[rk] returns a list of indices that should be
        ## sent from current process to process rk.
        ## Usage :  mpi.send(rk, ..., yourField[sendTo[rk]],...)
        self.sendTo = {}

        ## List of local indices that should be copied from data of
        ## topofrom to data of topoTo for the current process.
        ## Usage:
        ## field_topoto[ito] = field_topofrom[ifrom]
        ## --> no mpi messages
        self.ifrom = []
        ## List of local indices that should be updated on the
        ## current process for topoTo from values on the
        ## same process for topoFrom.
        ## Usage:
        ## field_topoto[ito] = field_topofrom[ifrom]
        ## --> no mpi messages
        self.ito = []

        self.hasLocalInter = True

        ## Parent communicator
        ## Todo : define some proper conditions for compatibility
        ## between topoFrom, topoTo and parent:
        ## - same size
        ## - same domain
        ## - common processus ...
        ## At the time we check that both topo have
        ## the same comm_origin.
        if parentComm is None:
            self._parentComm = self.topoFrom.parent()
        else:
            self._parentComm = parentComm
        self._parentRank = self._parentComm.Get_rank()
        self._parentSize = self._parentComm.Get_size()
        if self.topoFrom == self.topoTo:
            return

        assert(self.topoTo.parent() is self._parentComm)

        # Both topologies must have the
        # same number of mpi processes.
        assert topoFrom.size == topoTo.size

        dom = topoFrom.domain
        
        # 1 - Collect global mesh indices for each topology
        iglobFrom = topotools.collectGlobalIndices(topoFrom, self._parentComm)
        iglobTo = topotools.collectGlobalIndices(topoTo, self._parentComm)
        
        # Connectivity :
        sTo = np.zeros((0, 1 + dom.dimension * 2), dtype=np.int32)
        indexFrom = []
        for d in range(dom.dimension):
            indexFrom.append(range(iglobFrom[self._parentRank, d * 2],
                                   iglobFrom[self._parentRank, d * 2 + 1] + 1))

        line = np.zeros((1 + 2 * dom.dimension), dtype=np.int32)
        listRanks = [i for i in range(self._parentSize)
                     if i != self._parentRank]

        for i in listRanks:
            line[0] = i
            hasInter = True
            for d in range(dom.dimension):
                indexTo = range(iglobTo[i, d * 2], iglobTo[i, d * 2 + 1] + 1)
                interRow = [k for k in indexFrom[d] if k in indexTo]
                interRow.sort()
                if interRow.__len__():
                    line[d * 2 + 1] = interRow[0]
                    line[d * 2 + 2] = interRow[-1]
                else:
                    hasInter = False
                    break

            if hasInter:
                sTo = np.vstack((sTo, line))
        
        # Same operations but for current process
        localFrom = np.zeros((2 * dom.dimension), dtype=np.int32)
        for d in range(dom.dimension):
            
            indexTo = range(iglobTo[self._parentRank, d * 2],
                            iglobTo[self._parentRank, d * 2 + 1] + 1)
          
            interRow = [k for k in indexFrom[d] if k in indexTo]
            interRow.sort()
            if interRow.__len__():
                localFrom[d * 2] = interRow[0]
                localFrom[d * 2 + 1] = interRow[-1]
            else:
                self.hasLocalInter = False
                break

        if self.hasLocalInter:
            localTo = topoTo.toIndexLocal(localFrom)
            localFrom = topoFrom.toIndexLocal(localFrom)
            for d in range(dom.dimension):
                self.ifrom.append(slice(localFrom[2 * d],
                                        localFrom[2 * d + 1] + 1))
                self.ito.append(slice(localTo[2 * d],
                                      localTo[2 * d + 1] + 1))

        # --- Compute globalConnectivity on process 0 and distribute it ---
        # Global Connectivity : each line corresponds to a message :
        # globalConnectivity[i,:] = [i, j, 0, 3, 2, 6]
        # means that process i must send to process j its mesh
        # from node 0 to 3 in the first direction and from node 2 to
        # 6 in the second dir AND that process j must receive the same mesh
        # from process i. Nodes number given in the global index set.
        # Warning : process ranks are given in parentComm.
        globalConnectivity = np.zeros((sTo.shape[0], 2 +
                                       dom.dimension * 2), dtype=np.int32)

        globalConnectivity[:, 1:] = sTo.copy(order='A')
        globalConnectivity[:, 0] = self._parentRank

        if self._parentRank == 0:
            for i in range(1, self._parentSize):
                temp = self._parentComm.recv(source=i)
                globalConnectivity = np.vstack((globalConnectivity, temp))

        else:
            self._parentComm.ssend(globalConnectivity, dest=0)

        ## recvFrom[:,0] rank of the sending process in parentComm
        ## recvFrom[:,1:] local indices (start,end, ...)
        ## of the points to be receive for each direction.
        ## Example (in 2D)
        ## recvFrom = [[1, 2, 5, 1, 3]]
        ## means that the current process must receive from process 1
        ## an array that will be saved at positions of indices 2:5
        ## in the first dir and 1:3 in second dir
        rFrom = np.empty((), dtype=np.int32)
        if self._parentRank == 0:
            for rk in range(self._parentSize):
                cond1 = globalConnectivity[:, 1] == rk
                cond2 = globalConnectivity[:, 0] != rk
                cond = cond1 & cond2
                # We keep lines of globalConnectivity that
                # correspond to sendings to rk
                sol = np.compress(cond, globalConnectivity, axis=0)
                # We remove first column from which all elements
                # must be equal to rk
                cond1 = np.ones((sol.shape[1]), dtype=np.bool)
                cond1[1] = False
                sol = np.compress(cond1, sol, axis=1)
                # Send to rk the resulting matrix
                # Each line of this matrix tells that
                # sol[i, 0] process must send to rk
                # elements indicated by indices in sol[i, 1:]
                if(rk != self._parentRank):
                    self._parentComm.ssend(sol, dest=rk)
                else:
                    rFrom = sol.copy()

        else:
            rFrom = self._parentComm.recv(source=0)

        # Final setup for recvFrom and sendTo : shift from
        # global indices to local ones.
        for row in range(rFrom.shape[0]):
            rFrom[row, 1:] = topoTo.toIndexLocal(rFrom[row, 1:])
        for row in range(sTo.shape[0]):
            sTo[row, 1:] = topoFrom.toIndexLocal(sTo[row, 1:])

        row = 0
        for rk in rFrom[:, 0]:
            self.recvFrom[rk] = []
            for d in range(dom.dimension):
                self.recvFrom[rk].append(slice(rFrom[row, 2 * d + 1],
                                               rFrom[row, 2 * d + 2] + 1))
            row += 1

        row = 0
        for rk in sTo[:, 0]:
            self.sendTo[rk] = []
            for d in range(dom.dimension):
                self.sendTo[rk].append(slice(sTo[row, 2 * d + 1],
                                             sTo[row, 2 * d + 2] + 1))
            row += 1
            
    @debug
    def setUp(self):
        """

        """

    def __str__(self):
        """ TopologyBridge info display """
        s = '======== Bridge from topology ' + str(self.topoFrom.getId())
        s += ' to topology ' + str(self.topoTo.getId()) + ' ========\n'
        s += '[' + str(self._parentRank) + '] sendTo :'
        s += str(self.sendTo) + '\n'
        s += '[' + str(self._parentRank) + '] recvFrom :'
        s += str(self.recvFrom) + '\n'
        s += '[' + str(self._parentRank) + '] iTo :' + str(self.ito) + '\n'
        s += '[' + str(self._parentRank) + '] iFrom :' + str(self.ifrom) + '\n'
        s += '=================================\n'
        return s


class Bridge_intercomm(object):
    """
    This bridge is defined as the Bridge class.

    First, an intercommunicator is created from the tasks identifiers (used
    by user to split the parent communicator).

    The mesh indices are exchanged by process across the intercommunicator.
    Intersections of mesh are computed and indices slices are stored in the
    transfers dictionnary.

    """
    @debug
    def __init__(self, topo, parent_comm, id_from, id_to, proc_tasks):
        """
        Create a bridge between communicators of a single topology.
        @param topo : Topology to consider.
        @param parent_comm : MPI communicator that containts all process
        involved in this bridge.
        @param id_from : Identifier for the input task.
        @param id_to : Identifier for the output task.
        @param proc_task : Tasks assignment for process.
        """
        self.topo = topo
        self.id_from = id_from
        self.id_to = id_to
        self.parent_comm = parent_comm
        self.proc_tasks = proc_tasks
        parent_rank = self.parent_comm.Get_rank()
        my_comm = self.topo.comm
        my_rank = my_comm.Get_rank()
        dim = self.topo.domain.dimension

        # 0 - Create an intercommunicator
        # Create_intercomm attributes are:
        #   - local rank of leader process for current group (always 0)
        #   - parent communicator
        #   - rank of leader process in the remote group
        if self.proc_tasks[parent_rank] == self.id_from:
            self.inter_comm = my_comm.Create_intercomm(
                0,
                self.parent_comm, self.proc_tasks.index(self.id_to))
        if self.proc_tasks[parent_rank] == self.id_to:
            self.inter_comm = my_comm.Create_intercomm(
                0,
                self.parent_comm, self.proc_tasks.index(self.id_from))

        # 1 - Get global mesh indices of my domain
        my_start = self.topo.mesh.global_start
        my_end = self.topo.mesh.global_end
        my_slices = npw.zeros((dim * 2, ), dtype=np.int32)
        my_slices[0::2] = my_start
        my_slices[1::2] = my_end

        # 2 - Get global mesh indices of other domain
        if self.proc_tasks[parent_rank] == self.id_from:
            from_size = my_comm.Get_size()
            to_size = self.inter_comm.Get_remote_size()
            other_size = to_size
        if self.proc_tasks[parent_rank] == self.id_to:
            to_size = my_comm.Get_size()
            from_size = self.inter_comm.Get_remote_size()
            other_size = from_size

        # 3 - Get global mesh indices of other topology process
        # Each process broadcast his mesh indices to all process of other group
        # from_slices and to_slices are temporary arrays
        from_slices = npw.zeros((from_size, dim * 2), dtype=np.int32)
        to_slices = npw.zeros((to_size, dim * 2), dtype=np.int32)

        # 3.1 - each process of the 'from' group broadbast to the 'to' group
        for rk in xrange(from_size):
            if self.proc_tasks[parent_rank] == self.id_from:
                if my_rank == rk:
                    from_slices[rk, :] = my_slices
                    self.inter_comm.bcast(from_slices[rk, :],
                                          root=MPI.ROOT)
                else:
                    self.inter_comm.bcast(from_slices[rk, :],
                                          root=MPI.PROC_NULL)
            elif self.proc_tasks[parent_rank] == self.id_to:
                from_slices[rk, :] = self.inter_comm.bcast(from_slices[rk, :],
                                                           root=rk)
        # 3.2 - each process of the 'to' group broadbast to the 'from' group
        for rk in xrange(to_size):
            if self.proc_tasks[parent_rank] == self.id_to:
                if my_rank == rk:
                    to_slices[rk, :] = my_slices
                    self.inter_comm.bcast(to_slices[rk, :],
                                          root=MPI.ROOT)
                else:
                    self.inter_comm.bcast(to_slices[rk, :],
                                          root=MPI.PROC_NULL)
            elif self.proc_tasks[parent_rank] == self.id_from:
                to_slices[rk, :] = self.inter_comm.bcast(to_slices[rk, :],
                                                         root=rk)
        self.parent_comm.Barrier()

        if self.proc_tasks[parent_rank] == self.id_from:
            other_slices = to_slices
        elif self.proc_tasks[parent_rank] == self.id_to:
            other_slices = from_slices

        # 4. - Build connectivity
        self.transfers = {}
        for other_rk in xrange(other_size):
            intersect_slice = npw.zeros((dim * 2), dtype=np.int32)
            intersect_slice[0::2] = np.maximum(my_slices[0::2],
                                               other_slices[other_rk, 0::2])
            intersect_slice[1::2] = np.minimum(my_slices[1::2],
                                               other_slices[other_rk, 1::2])
            # Convert intersection to local slices, exluding ghosts
            if not (intersect_slice[0::2] > intersect_slice[1::2]).any():
                # compatible intersection (else nothing to transfer)
                self.transfers[other_rk] = [
                    (g + i - s, g + j + 1 - s) for i, j, s, g in zip(
                        intersect_slice[0::2],
                        intersect_slice[1::2], my_start, self.topo.ghosts)]

    @debug
    def setUp(self):
        """

        """

# ## Example of broadcast:
# ## Process 0 of group id_from broadcasting value to group id_to
# if proc_tasks[parent_rank] == id_from:
#     if my_rank == 0:
#         value = 123
#         inter_comm.bcast(value, root=MPI.ROOT)
#     else:
#         value = None
#         inter_comm.bcast(value, root=MPI.PROC_NULL)
#     value = comm_s.bcast(value, root=0) # for broad cast within the group
# elif proc_tasks[parent_rank] == id_to:
#     value = None
#     value = inter_comm.bcast(value, root=0)
# assert value == 123

# ## Example of Point-to-Point send/receve:
# x = npw.zeros((10, ))
# if my_rank == 0 and proc_tasks[parent_rank] == id_from:
#     x = npw.realarray(np.arange(10))
# if my_rank == 1 and proc_tasks[parent_rank] == id_to:
#     inter_comm.Recv([x, 10, PARMES_MPI_REAL], source=0, tag=0)
# if my_rank == 0 and proc_tasks[parent_rank] == id_from:
#     inter_comm.Send([x, 10, PARMES_MPI_REAL], dest=1, tag=0)
# if my_rank == 1 and proc_tasks[parent_rank] == id_to:
#     assert (x == npw.realarray(np.arange(10))).all()
