"""
@file topology.py

Parmes topologies definitions and related tools.
A Parmes Topology is defined as the association of
a mpi process distribution (mpi topology) and of a set of local meshes
(one per process).

At the time, only cartesian topologies with cartesian meshes are available.

This module provides the following classes :
- Cartesian : 1,2 or 3D parmes topology.
- Bridge : bridge between two parmes topologies : i.e. what must be
exchanged between two topologies and how.

To get more details try :
\code
>> import parmepy.mpi.topology as topo
>> help(topo.Cartesian)
>> help(topo.Bridge)
\endcode

"""

from parmepy.constants import ORDER, np, PARMES_INDEX, debug, PARMES_INTEGER
from parmepy.mpi.mesh import SubMesh
from itertools import count
from parmepy.mpi.main_var import MPI


class Cartesian(object):
    """
    Define a MPI cartesian topology with an associated
    mesh for each sub-domain.
    Keyword arguments :
    - domain : the geometry; it must be a box.
    - globalMeshResolution : user input. Number of points in the domain
    in each direction. We assume that first point corresponds to origin, and
    last point to boundary point, whatever the boundary type is.
    That is x[0] = domain.origin and
    x[globalMeshResolution-1] = domain.Lengths_x.
    - dim : dimension of the topology
    - shape : topology resolution (i.e process layout in each direction).
    - comm : MPI communicator used to create this topology
    - periods : periodicity of the topology (in each direction)
    - ghosts : number of points in the ghost layer

    """

    @debug
    def __new__(cls, *args, **kw):
        return object.__new__(cls, *args, **kw)

    # Counter of topology.Cartesian instances to set a unique id for each
    # Cartesian topology instance.
    __ids = count(0)

    @debug
    def __init__(self, domain, dim, globalMeshResolution,
                 comm=None, periods=None, ghosts=None, cutdir=None,
                 shape=None,
                 precomputed=False, localres=None, localoffset=None):

        assert domain.dimension == globalMeshResolution.__len__(), \
            'The resolution size differs from the domain dimension.'

        ## Associated domain
        self.domain = domain
        ## An id for the topology
        self.__id = self.__ids.next()
        ## Dim of the cartesian grid for MPI processes.
        self.dim = dim
        # Mind the sequential case : dim from withConstr may be equal
        # to 0 if shape[:] = 1
        ## (Source) Communicator used to build the topology
        if comm is None:
            from parmepy.mpi.main_var import main_comm as comm
        self._comm_origin = comm
        ## number of process in comm_origin
        self._origin_size = self._comm_origin.Get_size()
        ## True if everything has been properly set for this topology
        self.isUpToDate = False
        ## Grid of mpi process layout
        self.shape = np.ones(self.domain.dimension, dtype=PARMES_INDEX)
        # MPI processes layout.
        if shape is None:
            # If shape is not provided, computation of  the "optimal" processus
            # distribution for each direction of the grid topology.
            shape = np.asarray(MPI.Compute_dims(self._origin_size, self.dim),
                               dtype=PARMES_INDEX)

            if cutdir is not None:
                self.cutdir = np.array(cutdir, dtype=np.bool)
                self.shape[self.cutdir] = shape
            else:
                # Reorder shape according to the data layout
                # if arrays are in "fortran" order (column major)
                # the last dir has priority for distribution.
                # For C-like (row major) arrays, first dir is the
                # first to be distributed, and so on.
                self.shape[:self.dim] = shape
                self.shape.sort()
                if(ORDER == 'C'):
                    self.shape[:] = self.shape[::-1]
                self.cutdir = self.shape != 1
            # redim if necessary ...
            self.dim = self.shape[self.cutdir].size

        else:
            shape = np.asarray(shape)
            assert shape.size == self.domain.dimension, 'Input shape must be \
                of the same size as the domain dimension.'
            if cutdir is None:
                self.cutdir = shape != 1
            else:
                self.cutdir = cutdir

            self.shape = shape
            self.dim = self.shape[self.cutdir].size

        # Special care for the 1 process case:
        if self._origin_size == 1:
            self.dim = 1
            self.cutdir = shape != 1
            self.cutdir[-1] = True

        ## MPI process grid periodicity; default is true.
        self.periods = np.ones((self.dim), dtype=np.bool)
        if(periods is not None):
            assert (periods.size == self.dim)
            self.periods[:] = periods[:]
        self.comm = self._comm_origin.Create_cart(self.shape[self.cutdir],
                                                  periods=self.periods,
                                                  reorder=True)

        ## Size of the topology (i.e. total number of mpi processes)
        self.size = self.comm.Get_size()
        ## Rank of the current process in the topology
        self.rank = self.comm.Get_rank()
        ## Coordinates of the current process
        self.reduced_coords = np.asarray(self.comm.Get_coords(self.rank),
                                         dtype=PARMES_INDEX)
        ## Coordinates of the current process
        ## What is different between proc_coords and reduced_coords?
        ## --> proc_coords has values even for directions that
        ## are not distributed. If cutdir = [False, False, True]
        ## then reduced_coords = [ nx ] and proc_coords = [0, 0, nx]
        self.proc_coords = np.zeros(self.domain.dimension, dtype=PARMES_INDEX)
        self.proc_coords[self.cutdir] = self.reduced_coords
        ## Neighbours : self.neighbours[0,i] (resp. [1,i])
        ## previous (resp. next) neighbour in direction i
        ## (warning : direction in the grid of process).
        self.neighbours = np.zeros((2, self.dim), dtype=PARMES_INTEGER)
        for direction in range(self.dim):
            self.neighbours[:, direction] = self.comm.Shift(direction, 1)

        ## ghost layer (default = 0)
        if(ghosts is None):
            self.ghosts = np.zeros((self.domain.dimension),
                                   dtype=PARMES_INDEX)
        else:
            self.ghosts = np.asarray(ghosts, dtype=PARMES_INTEGER)
            assert(self.ghosts.size == domain.dimension)
        assert(np.all(self.ghosts >=0))
        self.hasGhosts = not np.all(self.ghosts == 0)

        # --- Computation of the local mesh resolution and indices ---
        ##  Resolution of the global mesh
        self.globalMeshResolution = np.asarray(globalMeshResolution,
                                               dtype=PARMES_INDEX)

        # If local resolution has already been computed
        # (for example from fftw init output)
        if precomputed:
            self.G_start = localoffset
            self.localGridResolution = localres

        # Usual case : we compute local resolution/offset
        else:
            # Number of "computed" points (i.e. excluding ghosts/boundaries).
            pts_noghost = np.zeros((self.domain.dimension), dtype=PARMES_INDEX)
            # Warning FP : we should test boundary conditions type here
            # If periodic, resol_calc = (globalMeshResolution - 1)
            # else, resol_calc =  (globalMeshResolution - 2)
            resolCalc = np.zeros((domain.dimension), dtype=PARMES_INDEX)
            resolCalc[:] = self.globalMeshResolution[:] - 1
            pts_noghost[:] = resolCalc // self.shape

            # If any, remaining points are
            # added on the mesh of the last process.
            remainingPts = np.zeros(self.domain.dimension, dtype=PARMES_INDEX)
            remainingPts[:] = resolCalc % self.shape

            # Then the total number of points (size of arrays to be allocated)
            nbpoints = pts_noghost.copy()
            for i in range(self.domain.dimension):
                if(self.proc_coords[i] == self.shape[i] - 1):
                    nbpoints[i] += remainingPts[i]

            self.localGridResolution = resolCalc.copy()
            self.localGridResolution[:] = nbpoints[:]
            self.localGridResolution[:] += 2 * self.ghosts[:]

            ## Global indices for the local mesh points;
            ## self.G_start[i] is the index of the first
            ## point in the i direction (It includes ghosts/boundary points).
            self.G_start = np.zeros((domain.dimension), dtype=PARMES_INDEX)
            self.G_start[:] = self.proc_coords[:] * pts_noghost[:]

        ## Local mesh on the current mpi process.
        self.mesh = None

        # The topology is register into domain list.
        # If it already exists (in the sense of the comparison operator
        # of the present class) then isNew is turned to false
        # and the present instance will be destroyed.
        self.isNew = True
        self.__id = self.domain.register(self)

    def parent(self):
        """
        returns the communicator used to build this topology
        """
        return self._comm_origin

    @debug
    def setUp(self):
        """Topology set up.

        Create topology's Local Mesh.
        This function is called during registration in the domain
        (domain.register(topo)).
        """
        if not self.isUpToDate:
            self.mesh = SubMesh(self, self.G_start, self.localGridResolution)
        self.isUpToDate = True

    @classmethod
    def withResolution(cls, domain, shape,
                       globalMeshResolution, ghosts=None, periods=None,
                       comm=None):
        """
        Compute a topology for a given shape of the processus grid.
        If the dimension of the topology is smaller than the domain dimension,
        shape will be reordered according to the following rule :
        (depends on array storage type (C or Fortran))
        - if C : first dir, first distributed.
        - if Fortran : last dir, first distributed.
        @param domain in which the topology is defined
        @param topoResolution : mpi process grid resolution (MUST be of
        the same size as domain), set topoResolution[i] = 1 if
        you don't want to cut the domain in direction i.
        @param ghosts : number of ghost points in each direction, default = 1.
        @param periods : periodicity in each direction, default = true.
        """

        shape = np.asarray(shape, dtype=PARMES_INDEX)
        assert shape.size == domain.dimension, 'Array for topo \
            resolution must be of size domain.dimension'
        dim = shape[shape != 1].size
        assert shape.all(), "Topology error : you try to \
            create a topology with 0 resolution in one direction."
        if dim < domain.dimension:
            shape.sort()
            if(ORDER == 'C'):
                shape[:] = shape[::-1]
        cutdir = shape != 1

        return cls(domain, dim, globalMeshResolution, cutdir=cutdir,
                   periods=periods, ghosts=ghosts, shape=shape,
                   comm=comm)

    @classmethod
    def withResolutionFixed(cls, domain, shape,
                            globalMeshResolution, ghosts=None, periods=None,
                            comm=None):
        """
        Compute a topology for a given shape of the processus grid.
        @param domain in which the topology is defined
        @param topoResolution : mpi process grid resolution (MUST be of
        the same size as domain), set topoResolution[i] = 1 if
        you don't want to cut the domain in direction i.
        @param ghosts : number of ghost points in each direction, default = 1.
        @param periods : periodicity in each direction, default = true.
        """
        shape = np.asarray(shape, dtype=PARMES_INDEX)
        assert shape.size == domain.dimension, 'Array for topo \
            resolution must be of size domain.dimension'
        dim = shape[shape != 1].size
        assert shape.all(), "Topology error : you try to \
            create a topology with 0 resolution in one direction."
        cutdir = shape != 1
        return cls(domain, dim, globalMeshResolution,
                   cutdir=cutdir, periods=periods, ghosts=ghosts, shape=shape,
                   comm=comm)

    @classmethod
    def withCutdir(cls, domain, cutdir,
                   globalMeshResolution, ghosts=None, periods=None,
                   comm=None):
        """
        Compute a topology from a list of directions to be cut.
        @param domain in which the topology is defined
        @param cutdir : array of bool, cutdir[i] = true if mesh must be
        distributed in direction i.
        (cutdir must be of the same size as the domain).
        @param ghosts : number of ghost points in each direction, default = 1.
        @param periods : periodicity in each direction, default = true.
        """
        cutdir = np.asarray(cutdir, dtype=np.bool)
        dim = cutdir[cutdir].size
        return cls(domain, dim, globalMeshResolution,
                   cutdir=cutdir, periods=periods, ghosts=ghosts,
                   comm=comm)

    @classmethod
    def withPrecomputedResolution(cls, domain, shape,
                                  globalMeshResolution, localres,
                                  offset, ghosts=None, periods=None,
                                  comm=None):
        """
        Compute a topology for a given shape of the processus grid
        and a given local mesh resolution/offset (example : output from
        fftw init process).
        @param domain in which the topology is defined
        @param topoResolution : mpi process grid resolution (MUST be of
        the same size as domain), set topoResolution[i] = 1 if
        you don't want to cut the domain in direction i.
        @param localres : local mesh resolution
        @param offset : local offset (i.e. global index of first local mesh
        points)
        @param ghosts : number of ghost points in each direction, default = 1.
        @param periods : periodicity in each direction, default = true.
        """
        shape = np.asarray(shape, dtype=PARMES_INDEX)
        assert shape.size == domain.dimension, 'Array for topo \
            resolution must be of size domain.dimension'
        dim = shape[shape != 1].size
        assert shape.all(), "Topology error : you try to \
            create a topology with 0 resolution in one direction."
        cutdir = shape != 1

        return cls(domain, dim, globalMeshResolution,
                   precomputed=True, localres=localres, localoffset=offset,
                   periods=periods, ghosts=ghosts, cutdir=cutdir, shape=shape,
                   comm=comm)

    def __eq__(self, other):
        """
        Operator for comparison. Based on:
        @li globalMeshResolution
        @li localMeshResolution
        @li shape
        @li domain
        @param other : Topology to compare with.
        @todo completer le test dans parmepy/mpi/tests/test_topology.py
        """
        return np.equal(self.globalMeshResolution,
                        other.globalMeshResolution).all() and \
            np.equal(self.localGridResolution,
                     other.localGridResolution).all() and \
            np.equal(self.shape, other.shape).all() and \
            np.equal(self.ghosts, other.ghosts).all() and \
            self.domain == other.domain

    def __ne__(self, other):
        """
        Not equal operator.
        Seems to be required in addition to __eq__ to
        avoid 'corner-case' behaviors.
        @param other : Topology to compare with.
        """
        result = self.__eq__(other)
        if result is NotImplemented:
            return result
        return not result

    def __str__(self):
        """ Topology info display """
        s = '======== Topology id : ' + str(self.__id) + ' ========\n'
        if(self.rank == 0):
            s += str(self.dim) + 'D cartesian Topology of size ' \
                + str(self.shape) + '\n'
        s += 'Process of coordinates ' + str(self.proc_coords[:])
        s += ' and of ranks (topo/origin) ' + str(self.rank) + '/'
        s += str(self._comm_origin.Get_rank()) + '.\n'
        s += 'Neighbours coordinates : \n'
        s += str(self.neighbours) + '\n'
        s += 'Ghost layer : ' + str(self.ghosts[:]) + '\n'
        s += 'Global mesh resolution : ' + str(self.globalMeshResolution)
        s += '\n'
        s += str(self.mesh)
        s += '\n=================================\n'
        return s

    def getId(self):
        """
        @return the id of the present topology.
        This id is unique among all defined topologies.
        """
        return self.__id

    def toIndexLocal(self, glob_index):
        """
        Return the local mesh indices from
        their global indices.
        Ilocal = Iglobal - G_start + ghost
        @param glob_index an array of size domain.dim*2
        with glob_index = [start_dir1, end_dir1, start_dir2, end_dir2, ...]
        return an array of indices, same 'setup' as glob_index.
        """
        ilocal = np.asarray(glob_index).copy()
        ilocal[0::2] = ilocal[0::2] - self.G_start + self.ghosts
        ilocal[1::2] = ilocal[1::2] - self.G_start + self.ghosts
        return ilocal

    def toIndexGlobal(self, sl):
        """
        Return the global mesh indices from
        their local indices.
        Iglobal = Ilocal + G_start - ghost
        @param loc_index an array of size domain.dim*2
        with loc_index = [start_dir1, end_dir1, start_dir2, end_dir2, ...]
        return an array of indices, same 'setup' as loc_index.
        """
        g_start = self.mesh.global_start
        return [slice(sl[i].start + g_start[i], sl[i].stop + g_start[i])
                for i in xrange(self.domain.dimension)]


class topotools():

    @staticmethod
    def collectGlobalIndices(topo, comm):
        """
        Get global start/end indices on
        the current process for the local mesh
        @param topo : the concerned topology
        @param comm : a communicator that will be
        used to gather the global indices. Usually
        the parent of topo.
        @return iglob : a list of arrays.
        iglob[rk] = [start_0, end_0, start_1, end_1...]
        start_i/end_i = global start/end indices for mesh of
        process number rk in comm.
        """
        size = topo.size
        start = topo.mesh.global_start
        end = topo.mesh.global_end
        # communicator that owns the topology
        rank = comm.Get_rank()
        iglob = np.zeros((size, topo.domain.dimension * 2), dtype=np.int32)
        iglob[rank, 0::2] = start
        iglob[rank, 1::2] = end
        
        comm.Allgather([iglob[rank, :], MPI.INT], [iglob, MPI.INT])
        return iglob

    
