# -*- coding: utf-8 -*-
"""
@file operator/adapt_timestep.py

Definition of the adaptative time step according to the flow fields.

"""
from parmepy.constants import debug
from parmepy.methods_keys import TimeIntegrator, SpaceDiscretisation,\
    dtCrit
from parmepy.numerics.finite_differences import FD_C_4
from parmepy.operator.discrete.adapt_timestep import AdaptTimeStep_D
from parmepy.operator.continuous import Operator
from parmepy.fields.variable_parameter import VariableParameter
import parmepy.default_methods as default


class AdaptTimeStep(Operator):
    """
    The adaptative Time Step is computed according
    to the following expression :
    dt_adapt = min (dt_advection, dt_stretching, dt_cfl)
    """

    @debug
    def __init__(self, velocity, vorticity, dt_adapt, io_params=None,
                 time_range=None, lcfl=0.125, cfl=0.5, **kwds):
        """
        Create a timeStep-evaluation operator from given
        velocity and vorticity variables.

        Note : If cfl is None, the computation of the adaptative time step
        is only based on dt_advection and dt_stretching, taking the minimum
        value of the two.

        @param velocity field
        @param vorticity field
        @param dt_adapt : adaptative timestep variable
        @param time_range : [start, end] use to define a 'window' in which
        the current operator is applied. Outside start-end, this operator
        has no effect. Start/end are iteration numbers.
        Default = [2, endofsimu]
        """
        super(AdaptTimeStep, self).__init__(variables=[velocity, vorticity],
                                            **kwds)
        if self.method is None:
            self.method = default.ADAPT_TIME_STEP
        ## velocity variable (vector)
        self.velocity = velocity
        ## vorticity variable (vector)
        self.vorticity = vorticity
        ## adaptative time step variable ("variable" object)
        self.dt_adapt = dt_adapt
        assert isinstance(self.dt_adapt, VariableParameter)
        # Check if 'dt' key is present in dt_adapt dict
        assert 'dt' in self.dt_adapt.data
        self.io_params = io_params
        assert SpaceDiscretisation in self.method.keys()
        assert TimeIntegrator in self.method.keys()
        if not dtCrit in self.method.keys():
            self.method[dtCrit] = 'vort'

        self.input = [self.velocity, self.vorticity]
        self.output = [self.vorticity]
        self.time_range = time_range
        self.lcfl, self.cfl = lcfl, cfl

    def discretize(self):
        if self.method[SpaceDiscretisation] is FD_C_4:
            nbGhosts = 2
        else:
            raise ValueError("Unknown method for space discretization of the\
                time step-evaluation operator.")
        super(AdaptTimeStep, self)._standard_discretize(nbGhosts)

    @debug
    def setUp(self):
        self.discreteOperator =\
            AdaptTimeStep_D(self.discreteFields[self.velocity],
                            self.discreteFields[self.vorticity],
                            self.dt_adapt, method=self.method,
                            time_range=self.time_range,
                            io_params=self.io_params,
                            lcfl=self.lcfl,
                            cfl=self.cfl)

        self.discreteOperator.setUp()
        self._isUpToDate = True
