"""
@file operator/analytic.py

Analytic operator representation.
"""
from parmepy.constants import debug
from parmepy.operator.continuous import Operator


class Analytic(Operator):
    """
    Applies an analytic formula, given by user, on its fields.
    """

    @debug
    ## def __init__(self, variables, resolutions=None, formula=None,
    ##              ghosts=None, topo=None, doVectorize=False, method=None,
    ##              task_id=None, comm=None):
    def __init__(self, formula=None, doVectorize=False, **kwds):
        """
        Create an operator using an analytic formula to compute field(s).
        @param variables : list of fields on which this operator will apply.
        @param formula : the formula to be applied
        @param resolutions : list of resolutions (one per variable)
        @param topo : a predefined topology to discretize variables
        @param ghosts : number of points in the ghost layer
        @param method : method used (Used to specify GPU support)

        @remark : method seems useless but it is useful for GPU.
        """
        super(Analytic, self).__init__(**kwds)
        if formula is not None:
            ## A formula applied to all variables of this operator
            self.formula = formula
            for v in self.variables:
                v.setFormula(formula, doVectorize)
            self.doVectorize = doVectorize
        else:
            assert self.variables[0].formula is not None
            self.formula = self.variables[0].formula
            self.doVectorize = self.variables[0].doVectorize
            # Only one formula allowed per operator
            for v in self.variables:
                assert v.formula is self.formula

        self.output = self.variables

    def discretize(self):
        super(Analytic, self)._standard_discretize()
        
    def setUp(self):
        self._isUpToDate = True

    @debug
    def apply(self, simulation=None):
        assert simulation is not None, \
            "Missing simulation value for computation."
        # Calling for requirements completion
        for red in self.requirements:
            red.wait()
        for v in self.variables:
            topo = self.discreteFields[v].topology
            v.initialize(currentTime=simulation.time, topo=topo)

    def __str__(self):
        s = super(Analytic, self).__str__()
        if self.discreteOperator is None:
            s += "Formula = " + str(self.formula)
        return s + "\n"
