"""
@file operator/continuous.py

Interface common to all continuous operators.
"""
from abc import ABCMeta, abstractmethod
from parmepy.constants import debug
import numpy as np
from parmepy.tools.timers import Timer


class Operator(object):
    """
    Abstract base class for continuous operators.

    An operator is composed of :
    - a set of continuous variables (at least one)
    - a method which defined how it would be discretized/processed
    - a discrete operator : object build using the method
    and the discretized variables.

    To each variable a 'resolution' is associated, used
    to create a topology and a discrete field.
    See details in 'real' operators (derived classes) description.
    """
    __metaclass__ = ABCMeta

    @debug
    def __new__(cls, *args, **kw):
        return object.__new__(cls, *args, **kw)

    @debug
    @abstractmethod
    def __init__(self, variables, method=None, topo=None, ghosts=None,
                 task_id=None, comm=None, resolutions=None):
        """
        Build the operator.
        The only required parameter is a list of variables.
        @param variables : list of fields on which this operator will apply.
        @param method : a dictionnary of methods.
        See methods.py for authorized values.
        the operator.
        @param name : operator id.
        @param topo : a predefined topology to discretize variables
        @param ghosts : number of points in the ghost layer
        @param comm : MPI communicator to build topologies
        Warning : you cannot set both topo and ghosts.
        @param resolutions : dictionnary of resolutions, one for each variable.
        Warning : you cannot set both topo and resolutions but you must set
        one of them, except for redistribute operators.
        """
        ## List of parmepy.continuous.Fields involved in the operator.
        if isinstance(variables, list):
            self.variables = variables
        else:
            self.variables = [variables]
        ## Input variables.
        self.input = []
        ## Output variables.
        self.output = []
        ## Domain of definition.
        self.domain = self.variables[0].domain
        for v in self.variables:
            assert v.domain is self.domain, 'All variables of the operator\
                must be defined on the same domain.'
        ## Object to describe the method of discretization of this operator.
        self.discreteOperator = None
        ## bool to check if the setup function has been called for
        ## this operator
        self._isUpToDate = False
        ## A dictionnary of discreteFields associated with this operator
        ## key = continuous variable \n
        ## Example : discrete_variable = discreteFields[velocity]
        ## returns the discrete fields that corresponds to the continuous
        ## variable velocity.
        self.discreteFields = {}
        ## The method used to discretize the operator.
        self.method = method
        ## Number of points in the ghost layer
        if ghosts is not None:
            self.ghosts = np.asarray(ghosts)
            assert self.ghosts.size == self.domain.dimension
        else:
            self.ghosts = None

        ## A predefined communicator. May be None
        self._comm = comm
        # Check if only one of topo or comm is given by user
        assert comm is None or topo is None

        # check that either topo or resolutions has been given
        # not true for redistribute operators ...
        # Note (JM):  This assertion is not possible in multiscale operators
        # assert topo is not None or resolutions is not None
        #if topo is not None and resolutions is not None:
        #    msg = "You can not set both resolutions and topology arguments."
        #    raise ValueError(msg)
        ## Topology used for ALL the variables of the operator
        self.topology = None
        if topo is None:
            ## Grid resolution for each variable (dictionnary)
            self.resolutions = resolutions
        elif resolutions is None:
            # Use the topology for all variables
            self.topology = topo
            self.resolutions = {}
            if self.ghosts is not None:
                assert (self.ghosts == topo.ghosts).all(), 'topo.ghosts and\
                input for ghosts must not be different'
            else:
                self.ghosts = topo.ghosts
            for v in self.variables:
                self.resolutions[v] = self.topology.globalMeshResolution
        else:
            # topology and resolution are given. Let the specific operator
            # mix the two.
            self.resolutions = resolutions
            self.topology = topo
            if self.ghosts is not None:
                assert (self.ghosts == topo.ghosts).all(), 'topo.ghosts and\
                input for ghosts must not be different'

        self.name = self.__class__.__name__
        self.task_id = task_id
        self.timer = Timer(self)
        ## Redistribute operator list that we must wait for.
        self.requirements = []
        ## time monitoring
        self.time_info = None

    @staticmethod
    def getWorkLengths(method=None, domain_dim=None):
        """
        Return the length of working arrays lists required
        for the discrete operator.
        """
        assert method is None
        assert domain_dim is None
        return 0, 0

    def setWorks(self, rwork=None, iwork=None):
        if rwork is None:
            rwork = []
        if iwork is None:
            iwork = []
        self.discreteOperator.setWorks(rwork, iwork)

    @abstractmethod
    def discretize(self):
        """
        Build (or get) topologies required for this operator
        and discretize all its variables.
        """

    def _standard_discretize(self, nbGhosts=0):
        """
        This functions provides a standard way to discretize the operator,
        but some operators may need a specific discretization process.
        """
        # One topo for all fields ...
        if self.topology is not None:
            self._discretize_single_topo(nbGhosts)
        else:
            # ... or one topo for each field.
            self._discretize_multi_topo(nbGhosts)

    def _discretize_single_topo(self, nbGhosts=0):
        """
        Discretization of all fields with the same topo,
        predefined during initialization.
        """
        assert self.topology is not None
        assert (self.topology.ghosts >= nbGhosts).all()
        for v in self.variables:
            self.discreteFields[v] = v.discretize(self.topology)

    def _discretize_multi_topo(self, nbGhosts=0):
        """
        A topo is build for each field, standard default mpi distribution,
        with global mesh resolution given at init.
        """
        if self.ghosts is not None:
            self.ghosts[self.ghosts < nbGhosts] = nbGhosts
        else:
            self.ghosts = np.ones((self.domain.dimension)) * nbGhosts
        for v in self.variables:
            topo = self.domain.getOrCreateTopology(self.domain.dimension,
                                                   self.resolutions[v],
                                                   ghosts=self.ghosts,
                                                   comm=self._comm)
            self.discreteFields[v] = v.discretize(topo)

    def addRedistributeRequirement(self, red):
        self.requirements.append(red)

    def getRedistributeRequirement(self):
        return self.requirements

    @abstractmethod
    def setUp(self):
        """
        Last step of initialization. After this, the operator must be
        ready for apply call.

        Main step : setup for discrete operators.
        """

    @debug
    def finalize(self):
        """
        Memory cleaning.
        """
        if self.discreteOperator is not None:
            self.discreteOperator.finalize()
            self.timer = self.timer + self.discreteOperator.timer

    @debug
    def apply(self, simulation=None):
        """
        Apply this operator to its variables.
        @param simulation : object that describes the simulation
        parameters (time, time step, iteration number ...), see
        parmepy.problem.simulation.Simulation for details.
        """
        for req in self.requirements:
            req.wait()
        self.discreteOperator.apply(simulation)

    def printComputeTime(self):
        """ Time monitoring."""
        if self.discreteOperator is not None:
            self.discreteOperator.printComputeTime()
            self.time_info = self.discreteOperator.time_info
        else:
            from parmepy.mpi.main_var import main_rank
            shortName = str(self.__class__).rpartition('.')[-1][0:-2]
            s = '[' + str(main_rank) + '] ' + shortName
            s += " : operator not discretized --> no computation, time = 0."
            print (s)

    def isUp(self):
        """
        True if ready to be applied (--> setup function has
        been called succesfully)
        """
        return self._isUpToDate

    def __str__(self):
        """
        Common printings for operators
        """
        shortName = str(self.__class__).rpartition('.')[-1][0:-2]
        if self.discreteOperator is not None:
            s = str(self.discreteOperator)
        else:
            s = shortName + " operator. Not discretised."
        return s + "\n"

    def updateGhosts(self):
        """
        Update ghost points values, if any.
        """
        self.discreteOperator.updateGhosts()


class EmptyOperator(object):
    """Empty operator"""
    def __init__(self, *args, **kwargs):
        pass

    def apply(self, *args, **kwargs):
        pass
