"""
@file operator/differential.py

Differential operators
"""
from parmepy.constants import debug, np
from parmepy.operator.continuous import Operator
from parmepy.operator.discrete.differential import CurlFFT, CurlFD, GradFD
from parmepy.methods_keys import SpaceDiscretisation
from parmepy.numerics.finite_differences import FD_C_4, FD_C_2
try:
    from parmepy.f2py import fftw2py
except ImportError:
    from parmepy.fakef2py import fftw2py
import parmepy.default_methods as default
from abc import ABCMeta, abstractmethod


class Differential(Operator):
    __metaclass__ = ABCMeta

    ## @debug
    ## def __new__(cls, *args, **kw):
    ##     return object.__new__(cls, *args, **kw)

    @debug
    def __init__(self, invar, outvar, **kwds):
        super(Differential, self).__init__(variables=[invar, outvar], **kwds)
        if self.method is None:
            self.method = default.DIFFERENTIAL

        ## input variable
        self.invar = invar
        ## Curl of input
        self.outvar = outvar

        if self.method[SpaceDiscretisation] is 'fftw':
            self.resolution = self.resolutions[self.outvar]
            assert self.resolution == self.resolutions[self.invar],\
                'for fftw method, all variables must have\
                the same global resolution.'
        self.output = [outvar]
        self.input = [invar]

    def discretize(self):

        domdim = self.domain.dimension
        # Get main communicator
        if self.topology is not None:
            comm = self.topology.comm
            commsize = self.topology.size
        elif self._comm is not None:
            comm = self._comm
            commsize = self._comm.Get_size()
        else:
            from parmepy.mpi.main_var import main_comm as comm
            from parmepy.mpi.main_var import main_size as commsize

        if self.method[SpaceDiscretisation] is 'fftw':
            # FFTW case : init fftw (fortran) solver
            localres, localoffset = fftw2py.init_fftw_solver(
                self.resolution, self.domain.length, comm=comm.py2f())
            topodims = np.ones(domdim)
            topodims[-1] = commsize

            # Case 1 : topology provided by user at init:
            if self.topology is not None:
                # Check if input topo is complient with fftw topo
                assert (self.topology.shape == topodims).all(), 'input topology is\
                        not compliant with fftw.'
                self._discretize_single_topo()

            else:
                for v in self.variables:
                    topo = self.domain.getOrCreateTopology(
                        self.domain.dimension,
                        self.resolution, topodims,
                        precomputed=True,
                        offset=localoffset,
                        localres=localres,
                        ghosts=self.ghosts,
                        comm=self._comm)
                    self.discreteFields[v] = v.discretize(topo)

        elif str(self.method[SpaceDiscretisation]).find('FD_C') != -1:
            # Finite differences method
            # Minimal number of ghost points
            if self.method[SpaceDiscretisation] is FD_C_4:
                nbGhosts = 2
            elif self.method[SpaceDiscretisation] is FD_C_2:
                nbGhosts = 1

            # Case 1 : topology provided by user at init:
            if self.topology is not None:
                self._discretize_single_topo(nbGhosts)

            else:
                if self.ghosts is None:
                    self.ghosts = np.asarray([nbGhosts] * domdim)
                else:
                    assert (self.ghosts >= nbGhosts).all()

                for v in self.variables:
                    topo = self.domain.getOrCreateTopology(domdim,
                                                           self.resolutions[v],
                                                           ghosts=self.ghosts,
                                                           comm=self._comm)
                    self.discreteFields[v] = v.discretize(topo)

        else:
            raise ValueError("Unknown method for space discretization of the\
                differential operator.")

        assert self.discreteFields[self.invar].topology == \
            self.discreteFields[self.outvar].topology, \
            'Operator not yet implemented for multiple resolutions.'

    @abstractmethod
    def setUp(self):
        """
        Last step of initialization. After this, the operator must be
        ready for apply call.

        Main step : setup for discrete operators.
        """


class Curl(Differential):
    """
    Computes \f$ outVar = \nabla inVar \f$
    """

    @debug
    def setUp(self):
        # Create a discrete operator, according to the chosen method
        assert self.domain.dimension == 3, "Not yet implemented for dim < 3"
        # todo : implement a 'false' 2D curl.
        if self.method[SpaceDiscretisation] is 'fftw':
            self.discreteOperator = CurlFFT(self.discreteFields[self.invar],
                                            self.discreteFields[self.outvar],
                                            method=self.method)
        elif str(self.method[SpaceDiscretisation]).find('FD_C') != -1:
            self.discreteOperator = CurlFD(self.discreteFields[self.invar],
                                           self.discreteFields[self.outvar],
                                           method=self.method)
        else:
            raise ValueError("The required Space Discretisation is\
                not available for Curl.")
        self.discreteOperator.setUp()
        self._isUpToDate = True

    @staticmethod
    def getWorkLengths(method=None, domain_dim=None):
        """
        Return the length of working arrays lists required
        for curl discrete operator
        @param method: useless, just there to fit with base class interface.
        @param domain_dim : useless, just there
        to fit with base class interface.
        """
        if method[SpaceDiscretisation] is FD_C_4:
            return CurlFD.getWorkLengths()
        else:
            return 0

    def setWorks(self, rwork=None, iwork=None):
        self.discreteOperator.setWorks(rwork, iwork)


class Grad(Differential):
    """
    Computes \f$ outVar = \nabla inVar \f$
    """

    @debug
    def setUp(self):
        if str(self.method[SpaceDiscretisation]).find('FD_C') == -1:
            raise ValueError("Grad operator only\
                implemented with finite differences. Please change\
                method[SpaceDiscretisation] value.")
        # Create a discrete operator, according to the chosen method
        # (only finite differences at the time).
        self.discreteOperator = GradFD(self.discreteFields[self.invar],
                                       self.discreteFields[self.outvar],
                                       method=self.method)
        self.discreteOperator.setUp()
        self._isUpToDate = True
