# -*- coding: utf-8 -*-
"""
@file diffusion.py

Operator for diffusion problem.

"""
from parmepy.operator.continuous import Operator
try:
    from parmepy.f2py import fftw2py
except ImportError:
    from parmepy.fakef2py import fftw2py
from parmepy.operator.discrete.diffusion_fft import DiffusionFFT
from parmepy.constants import debug
from parmepy.methods_keys import SpaceDiscretisation
import numpy as np
import parmepy.default_methods as default


class Diffusion(Operator):
    """
    Diffusion operator
    \f{eqnarray*}
    \omega = Op(\omega)
    \f} with :
    \f{eqnarray*}
    \frac{\partial \omega}{\partial t} &=& \nu\Delta\omega
    \f}
    """

    @debug
    def __init__(self, vorticity, viscosity, **kwds):
        """
        Constructor for the diffusion operator.
        @param[in,out] vorticity : field \f$ \omega \f$
        @param[in] resolution :  \f$ \omega \f$ global resolution.
        @param[in] viscosity : \f$\nu\f$, viscosity of the considered medium.
        """
        super(Diffusion, self).__init__(variables=[vorticity], **kwds)
        # The only available method at the time is fftw
        if self.method is None:
            self.method = default.DIFFUSION

        ## input/output field, solution of the problem
        self.vorticity = vorticity
        ## viscosity
        self.viscosity = viscosity

        self.input = [self.vorticity]
        self.output = [self.vorticity]

    def discretize(self):
        # The only available solver is fftw
        if self.method[SpaceDiscretisation] is not 'fftw':
            print (self.method)
            raise AttributeError("Method not yet implemented.")

        # Compute local resolution/distribution of data
        # according to fftw requirements.
        if self.topology is not None:
            comm = self.topology.comm
        elif self._comm is not None:
            comm = self._comm
        else:
            from parmepy.mpi.main_var import main_comm as comm
        localres, localoffset = fftw2py.init_fftw_solver(
            self.resolutions[self.vorticity],
            self.domain.length, comm=comm.py2f())

        if self.topology is not None:
            main_size = self.topology.size
        elif self._comm is not None:
            main_size = self._comm.Get_size()
        else:
            from parmepy.mpi.main_var import main_size
        topodims = np.ones((self.domain.dimension))
        topodims[-1] = main_size
        #variables discretization
        if self.ghosts is not None:
            raise AttributeError("Ghosts points not yet\
            implemented for diffusion operator.")
        if self.topology is not None:
            assert (self.topology.shape == topodims).all(), 'input topology is\
                    not compliant with fftw.'
            for v in self.variables:
                self.discreteFields[v] = v.discretize(self.topology)
        else:
            for v in self.variables:
                topo = self.domain.getOrCreateTopology(
                    self.domain.dimension,
                    self.resolutions[self.vorticity], topodims,
                    precomputed=True,
                    offset=localoffset,
                    localres=localres,
                    ghosts=self.ghosts,
                    comm=self._comm)
                self.discreteFields[v] = v.discretize(topo)

    @debug
    def setUp(self):
        """
        Diffusion operator setup : discretization of fields and operator.
        """
        self.discreteOperator = DiffusionFFT(
            self.discreteFields[self.vorticity], self.viscosity,
            method=self.method)

        self.discreteOperator.setUp()
        self._isUpToDate = True
