# -*- coding: utf-8 -*-
"""
@file operator/discrete/velocity_correction.py

Correction of the velocity field.
"""

from parmepy.constants import debug
from parmepy.operator.discrete.discrete import DiscreteOperator
from parmepy.fields.variable_parameter import VariableParameter
from parmepy.domain.obstacle.controlBox import ControlBox
from parmepy.tools.timers import timed_function
import parmepy.tools.numpywrappers as npw
from parmepy.constants import np, XDIR, YDIR, ZDIR
from parmepy.mpi import MPI


class VelocityCorrection_D(DiscreteOperator):
    """
    The velocity field is corrected after solving the
    Poisson equation. For more details about calculations,
    see the "velocity_correction.pdf" explanations document
    in ParmeDoc directory.
    """

    @debug
    def __init__(self, velocity, vorticity, req_flowrate, cb):
        """
        @param[in, out] velocity field to be corrected
        @param[in] vorticity field used to compute correction
        @param[in] req_flowrate : required value for the 
        flowrate (VariableParameter object)
        @param[in] surf : surface (parmepy.domain.obstacle.planes.SubPlane)
        used to compute reference flow rates. Default = surface at x_origin,
        normal to x-dir.
        """
        ## velocity discrete field
        self.velocity = velocity
        ## vorticity discrete field
        self.vorticity = vorticity
        ## domain dimension
        self.dim = self.velocity.domain.dimension
        # If 2D problem, vorticity must be a scalar
        if self.dim == 2:
            assert self.vorticity.nbComponents == 1
        assert (self.dim == 2 or self.dim == 3),\
            "Wrong problem dimension: only 2D and 3D cases are implemented."

        DiscreteOperator.__init__(self, [self.velocity, self.vorticity])

        self.input = [self.velocity, self.vorticity]
        self.output = [self.velocity]
        ## Velocity from topology will be used as reference
        self.topo = self.velocity.topology
        hh = self.topo.mesh.space_step
        self.cb = cb
        self.cb.discretize(self.topo)
        self.surfRef = cb.lowerS[XDIR]
        normal = self.surfRef.normal
        dirs = np.where(normal == 0)[0]
        self.ds = np.prod(self.surfRef.lengths[dirs] + hh[dirs])

        ## Expected value for the flow rate through self.surfRef
        self.req_flowrate = req_flowrate
        assert isinstance(self.req_flowrate, VariableParameter),\
            "the required flowrate must be a VariableParameter object."
        ## The correction that must be applied on each
        ## component of the velocity.
        self.velocity_shift = npw.zeros(self.dim)
        nbf = self.velocity.nbComponents + self.vorticity.nbComponents
        # temp buffer, used to save flow rates and mean
        # values of vorticity
        self.rates = npw.zeros(nbf)
        self.req_flowrate_val = None

        # Open file for output
        import parmepy.tools.io_utils as io
        io_default={"filename":'om_bar'}
        self.fname = io.Writer(params=io_default).filename
        if (self.topo.rank == 0):
            self.f = open(self.fname, 'w')

    def setUp(self):
        spaceStep = self.topo.mesh.space_step
        lengths = self.topo.domain.length
        self.coeff_mean = npw.prod(spaceStep) / npw.prod(lengths)
        x0 = self.surfRef.origin[XDIR]
        self.x_coord = self.topo.mesh.coords[XDIR] - x0
        self._isUpToDate = True

    def computeCorrection(self):
        """
        Compute the required correction for the current state
        but do not apply it onto velocity.
        """
        ## Computation of the flowrates evaluated from
        ## current (ie non corrected) velocity
        ## local flow reduced only on proc 0
        nbf = self.velocity.nbComponents + self.vorticity.nbComponents
        localrates = npw.zeros((nbf))
        for i in xrange(self.velocity.nbComponents):
            localrates[i] = self.velocity.integrateOnSurf_proc(self.surfRef,
                                                               component=i)
        start = self.velocity.nbComponents
        ## Integrate vorticity over the whole domain
        for i in xrange(self.vorticity.nbComponents):
            localrates[start + i] = \
                self.vorticity.integrate_on_proc(self.cb, component=i)

        # MPI reduction for rates
        # rates = [flowrate[X], flowrate[Y], flowrate[Z],
        #          vort_mean[X], ..., vort_mean[Z]]
        # or (in 2D) = [flowrate[X], flowrate[Y], vort_mean]
        self.rates[...] = 0.0
        self.velocity.topology.comm.Allreduce(localrates, self.rates)

        dlengths = npw.prod(self.cb.lengths)
        for i in xrange(self.vorticity.nbComponents):
            self.rates[start + i] /= dlengths

        # Set velocity_shift == [Vx_shift, vort_mean[Y], vort_mean[Z]]
        # or (in 2D) velocity_shift == [Vx_shift, vort_mean]
        # Velocity shift for main dir component
        self.velocity_shift[XDIR] = self.req_flowrate_val[XDIR]\
            - self.rates[XDIR] / self.ds
        # Shifts in other directions depend on x coord
        # and will be computed during apply.

    @timed_function
    def apply(self, simulation=None):
        # Calling for requirements completion
        DiscreteOperator.apply(self, simulation)
        ctime = MPI.Wtime()

        # the required flowrate value is updated (depending on time)
        self.req_flowrate.update(simulation)

        # warning : the flow rate value is divided with area of input surf.
        self.req_flowrate_val = self.req_flowrate[self.req_flowrate.name] \
            / self.ds
        # Computation of the required velocity shift
        # for the current state
        self.computeCorrection()
        iCompute = self.topo.mesh.iCompute

        # Apply shift to velocity
        self.velocity[XDIR][iCompute] += self.velocity_shift[XDIR]
        start = self.velocity.nbComponents
        # reminder : self.rates =[vx_shift, flowrates[Y], flowrate[Z],
        #                         vort_mean[X], vort_mean[Y], vort_mean[Z]]
        # or (in 2D) [vx_shift, flowrates[Y], vort_mean]
        vort_mean = self.rates[start:]
        t = simulation.time
        ite = simulation.currentIteration
        if self.topo.rank == 0:
            self.f = open(self.fname, 'a')
            self.f.write("%s      %s      %s      %s      %s \n" % (t, 
                                                                    ite, 
                                                                    vort_mean[0], 
                                                                    vort_mean[1], 
                                                                    vort_mean[2]))
            self.f.close()

        if self.dim == 2:
            # Correction of the Y-velocity component
            self.velocity[YDIR][...] += self.req_flowrate_val[YDIR] + \
                vort_mean[XDIR] * self.x_coord - self.rates[YDIR] / self.ds

        elif self.dim == 3:
            # Correction of the Y and Z-velocity components
            self.velocity[YDIR][...] +=  self.req_flowrate_val[YDIR] + \
                vort_mean[ZDIR] * self.x_coord - self.rates[YDIR] / self.ds
            self.velocity[ZDIR][...] += self.req_flowrate_val[ZDIR] - \
                vort_mean[YDIR] * self.x_coord - self.rates[ZDIR] / self.ds

        self._apply_timer.append_time(MPI.Wtime() - ctime)

