# -*- coding: utf-8 -*-
"""
@file energy_enstrophy.py
Compute Energy and Enstrophy
"""
import numpy as np
from parmepy.constants import debug, XDIR
from parmepy.operator.monitors.monitoring import Monitoring
from parmepy.tools.timers import timed_function
import parmepy.tools.numpywrappers as npw


class Energy_enstrophy(Monitoring):
    """
    Computes enstrophy and the kinetic energy
    \f{eqnarray*}
    enstrophy = \frac{1}{\Omega}\int_\Omega \omega^2 d\Omega
    \f} with \f$\Omega\f$ the volume or surface of the physical domain
    \f$\omega\f$ the vorticity and
    \f{eqnarray*}
    energy = \frac{1}{2\Omega}\int_\Omega v^2 d\Omega
    \f}
    """

    def __init__(self, velocity, vorticity,
                 viscosity, isNormalized, **kwds):
        """
        Constructor.
        @param velocity field
        @param vorticity field
        @param viscosity : kinematic viscosity
        @param isNormalized : boolean indicating whether the enstrophy
        and energy values have to be normalized by the domain lengths.
        @param topo : the topology on which we want to monitor the fields
        @param io_params : parameters (dict) to set file output.
        If  None, no output. Set io_params = {} if you want output,
        with default parameters values. 
        Default file name = 'energy_enstrophy.dat'
        See parmepy.tools.io_utils.Writer for details
        """
        if 'io_params' in kwds:
            params = kwds['io_params']
            if not "filename" in params:
                params["filename"] = "energy_enstrophy"
            # Set output buffer shape
            params["writebuffshape"] = (1, 3)
        super(Energy_enstrophy, self).__init__(variables=[velocity, vorticity],
                                               **kwds)
        ## velocity field
        self.velocity = velocity
        ## vorticity field
        self.vorticity = vorticity
        ## viscosity (scalar)
        self.viscosity = viscosity
        ## are the energy end enstrophy values normalized by domain lengths ?
        self.isNormalized = isNormalized
        ## self._buffer_1 = 0.
        ## self._buffer_2 = 0.
        self.input = [velocity, vorticity]
        self.output = []
        # \todo : rewrite for multiresolution case.
        # Note FP : for multiresolution case, it would probably be
        # better to use two different operators for energy and enstrophy.

    def setUp(self):
        if not self._isUpToDate:
            self.discreteFields[self.velocity] = \
                self.velocity.discretization(self.topology)
            self.discreteFields[self.vorticity] =\
                self.vorticity.discretization(self.topology)

            spaceStep = self.topology.mesh.space_step
            length = self.topology.domain.length
            if self.isNormalized:
                self.coeffEnstrophy = (np.prod(spaceStep) /
                                       np.prod(length))
            else:
                self.coeffEnstrophy = np.prod(spaceStep)
            self.coeffEnergy = 0.5 * self.coeffEnstrophy

            # A work vector for local computations
            # Warning : we assume one topo for all variables
            self.ind = self.topology.mesh.iCompute
            shape = self.discreteFields[self.velocity].data[0][self.ind].shape
            self._work = npw.zeros(shape)
            self._isUpToDate = True

    @debug
    @timed_function
    def apply(self, simulation=None):
        """
        Computation of kinetic energy, enstrophy &
        Checking energy and enstrophy decay
        """
        if simulation is None:
            raise ValueError("Missing simulation value for computation.")

        #time = MPI.Wtime()

        # --- Kinetic energy computation ---
        vd = self.discreteFields[self.velocity]
        # get the list of computation points (no ghosts)
        nbc = vd.nbComponents
        # Integrate (locally) velocity ** 2
        self._work[...] = vd[XDIR][self.ind] ** 2
        [np.add(self._work[...], vd[i][self.ind] ** 2, self._work[...])
         for i in xrange(1, nbc)]
        local_energy = np.sum(self._work)

        # --- Enstrophy computation ---
        vortd = self.discreteFields[self.vorticity]
        nbc = vortd.nbComponents
        # Integrate (locally) vorticity ** 2
        self._work[...] = vortd[0][self.ind] ** 2
        [np.add(self._work[...], vortd[i][self.ind] ** 2, self._work[...])
         for i in xrange(1, nbc)]
        local_enstrophy = np.sum(self._work)

        # --- Reduce energy and enstrophy values overs all proc ---
        # two ways : numpy or classical. Todo : check perf and comm
        sendbuff = npw.zeros((2))
        recvbuff = npw.zeros((2))
        sendbuff[:] = [local_energy, local_enstrophy]
        self.topology.comm.Allreduce(sendbuff, recvbuff)
        # the other way :
        #energy = self._topovel.topo.allreduce(local_energy, PARMES_MPI_REAL,
        #                                     op=MPI.SUM)
        #enstrophy = self._topovel.topo.allreduce(local_enstrophy,
        #                                        PARMES_MPI_REAL,
        #                                        op=MPI.SUM)

        # Update global values
        energy = recvbuff[0] * self.coeffEnergy
        enstrophy = recvbuff[1] * self.coeffEnstrophy

        # Update buffers
        ## energyBuff1 = self._buffer_1
        ## energyBuff2 = self._buffer_2
        ## self._buffer_2 = self._buffer_1
        ## self._buffer_1 = energy

        # Print results, if required
        ite = simulation.currentIteration
        if self._writer is not None and self._writer.doWrite(ite):
            self._writer.buffer[0, 0] = simulation.time
            self._writer.buffer[0, 1] = energy
            self._writer.buffer[0, 2] = enstrophy
            self._writer.write()
