# -*- coding: utf-8 -*-
"""
@file poisson.py
Poisson problem.

"""
from parmepy.operator.continuous import Operator
try:
    from parmepy.f2py import fftw2py
except ImportError:
    from parmepy.fakef2py import fftw2py

from parmepy.operator.discrete.poisson_fft import PoissonFFT
from parmepy.constants import debug
import numpy as np
from parmepy.mpi.main_var import main_size
from parmepy.operator.velocity_correction import VelocityCorrection
from parmepy.methods_keys import SpaceDiscretisation
import parmepy.default_methods as default


class Poisson(Operator):
    """
    Poisson problem,
    \f{eqnarray*}
    v = Op(\omega)
    \f} with :
    \f{eqnarray*}
    \Delta \phi &=& -\omega \\
    v &=& \nabla \times \phi
    \f}
    """

    @debug
    def __init__(self, velocity, vorticity, flowrate=None, **kwds):
        """
        Constructor for the Poisson problem.

        @param[out] velocity : solution field
        @param[in] vorticity : rhs field
        @param[in] resolutions :  list of resolutions (one per variable)
        @param[in] ghosts : ghost layer (optional), default=[0, 0, ...]
        @param[in] method : numerical method (default = fft)
        @param[in] flowrate : a flow rate value (through input surf,
        normal to xdir) used to compute a correction of the solution field.
        Default = 0 (no correction). See parmepy.operator.velocity_correction.
        """
        ## solution of the problem
        self.velocity = velocity
        ## -(right-hand side)
        self.vorticity = vorticity
        # Warning : for fftw all variables must have
        # the same resolution.
        super(Poisson, self).__init__(variables=[velocity, vorticity], **kwds)

        if self.method is None:
            self.method = default.POISSON

        if self.method[SpaceDiscretisation] is 'fftw':
            self.resolution = self.resolutions[self.velocity]
            assert self.resolution == self.resolutions[self.vorticity],\
                'Poisson error : for fftw, all variables must have\
                the same global resolution.'
        else:
            raise AttributeError("Method not yet implemented.")

        self.input = [self.vorticity]
        self.output = [self.velocity]
        if flowrate is not None:
            self.withCorrection = True
            self._flowrate = flowrate
        else:
            self.withCorrection = False
        self.correction = None
        self.projection = None

    def discretize(self):
        # Poisson solver based on fftw
        if (self.ghosts is not None) and (np.sum(self.ghosts)!=0):
            raise AttributeError("Ghosts points not yet\
            implemented for poisson operator.")

        # Compute local resolution/distribution of data
        # according to fftw requirements.
        if self.topology is not None:
            comm = self.topology.comm
        elif self._comm is not None:
            comm = self._comm
        else:
            from parmepy.mpi.main_var import main_comm as comm
        localres, localoffset = fftw2py.init_fftw_solver(
            self.resolution, self.domain.length, comm=comm.py2f())

        if self.topology is not None:
            commsize = self.topology.size
        elif self._comm is not None:
            commsize = self._comm.Get_size()
        else:
            commsize = main_size
        topodims = np.ones((self.domain.dimension))
        topodims[-1] = commsize
        #variables discretization
        if self.topology is not None:
            assert (self.topology.shape == topodims).all(), 'input topology is\
                    not compliant with fftw.'
            self._discretize_single_topo()

        else:
            for v in self.variables:
                topo = self.domain.getOrCreateTopology(
                    self.domain.dimension,
                    self.resolution, topodims,
                    precomputed=True,
                    offset=localoffset,
                    localres=localres,
                    ghosts=self.ghosts,
                    comm=self._comm)
                self.discreteFields[v] = v.discretize(topo)
        # Build and discretize correction operator, if necessary
        if self.withCorrection:
            self.correction = VelocityCorrection(self.velocity, self.vorticity,
                                                 resolutions=self.resolutions,
                                                 flowrate=self._flowrate,
                                                 topo=self.velocity.topology)
            self.correction.discretize()

    @debug
    def setUp(self):
        """
        """
        # Build and setup for the discrete operator
        if self.withCorrection:
            self.correction.setUp()
            cd = self.correction.discreteOperator
        else:
            cd = None
        self.discreteOperator = PoissonFFT(self.discreteFields[self.velocity],
                                           self.discreteFields[self.vorticity],
                                           correction=cd,
                                           projection=self.projection)
        self.discreteOperator.setUp()
        self._isUpToDate = True

    def activateProjection(self, projection):
        # If there is a projection, vorticity is also an output
        self.output.append(self.vorticity)
        self.projection = projection
