"""
@file redistribute.py
Setup for data transfer/redistribution between two parmes topologies.

This operator is an inter operator which is supposed to define the process to
transfer variables and data from one operator to another.
This mainly concerns data redistribution if the two operators work on
different mpi topologies.

When is it required to define an operator between op1 and op2?
If:
- the intersection between op1.output-variables and op2.input-variables
  is not empty
- AND if the topology on which the variables of op1 are defined is different
from the one of op2 variables.

Note Franck: this kind of operator may also be useful
to define the interpolation/filter process for data transfer for
a variable defined on several meshes.

"""
from parmepy import __VERBOSE__
from parmepy.constants import debug, PARMES_MPI_REAL, ORDERMPI, np, S_DIR
from parmepy.operator.continuous import Operator
from parmepy.mpi.bridge import Bridge
from parmepy.methods_keys import Support


class Redistribute(Operator):
    """
    Interconnection between two operators.
    SetUp will compute (or get if it already exists) a Bridge between two
    topologies.
    Apply redistributes data from opFrom topology to opTo topology.

    """
    @debug
    def __init__(self, opFrom, opTo, name_suffix=None, component=None, **kwds):

        """
        Create an operator to distribute data between two mpi topologies for a
        list of variables belonging to two operators.

        @param variables : the set of variables to be redistributed
        @param opFrom : source operator
        @param opTo : target (i.e.) the operator that handles the topology on
        which data must be redistributed.
        @param component: components of vector fields to consider (default:
        None, all components are taken).
        """
        super(Redistribute, self).__init__(**kwds)
        vars_str = "_("
        for vv in self.variables:
            vars_str += vv.name + ","
        vars_str = vars_str[:-1] + ')'
        if component is not None:
            vars_str += S_DIR[component]
        if name_suffix is None:
            name_suffix = ''
        self.name += vars_str + name_suffix
        ## Source Operator
        self.opFrom = opFrom
        ## Targeted operator
        self.opTo = opTo

        self.input = self.output = self.variables
        self.evts = []
        self._toHost_fields = []
        self._toDevice_fields = []
        self._hasRequests = False
        self.component = component
        if component is None:
            # All components are considered
            self._range_components = lambda v: range(v.nbComponents)
        else:
            # Only the given component is considered
            self._range_components = lambda v: [component]
        self.r_request = {}
        self.s_request = {}
        self._r_types = {}
        self._s_types = {}
        for v in self.variables:
            self._r_types[v] = {}
            self._s_types[v] = {}

        # Enable desynchronization: the opTo operator must call the wait
        # function of this redistribute. This operator has to know self.
        self.opTo.addRedistributeRequirement(self)

    def discretize(self):
        pass

    @debug
    def setUp(self):
        """
        Computes intersection of two topologies.

        """
        # Then check if variables belong to both operators
        # And check if variables have enought components.
        for v in self.variables:
            assert v in self.opFrom.variables and v in self.opTo.variables, \
                'Redistribute error : one of the variable is not present\
                in both source and target operator.'
            if self.component is not None:
                assert self.component >= 0, 'component needs to be positive'
            assert v.nbComponents > self.component, \
                'Redistribute error : variable ' + str(v.name) + ' do not \
                have enough components (' + str(self.component) + ')'
        assert self.opFrom.isUp() and self.opTo.isUp(), \
            """You should setup both opFrom and opTo operators
            before any attempt to setup a redistribute operator."""

        # Look for an operator operating on device.
        try:
            opFrom_is_device = \
                self.opFrom.method[Support].find('gpu') >= 0
        except KeyError:  # op.method is a dict not containing Support in keys
            opFrom_is_device = False
        except IndexError:  # op.method is a string
            opFrom_is_device = False
        except TypeError:  # op.method is None
            opFrom_is_device = False
        try:
            opTo_is_device = \
                self.opTo.method[Support].find('gpu') >= 0
        except KeyError:  # op.method is a dict not containing Support in keys
            opTo_is_device = False
        except IndexError:  # op.method is a sting
            opTo_is_device = False
        except TypeError:  # op.method is None
            opTo_is_device = False

        if not opFrom_is_device and not opTo_is_device:
            # case: opFrom(host) --host--> opTo(host)
            self.apply = self._host
            self.wait = self._wait_host
        else:
            # Have on device operators
            self.wait = self._wait_all
            if opFrom_is_device and not opTo_is_device:
                # case: opFrom(GPU) --toHost--host--> opTo(host)
                self.apply = self._apply_toHost_host
            elif not opFrom_is_device and opTo_is_device:
                # case: opFrom(host) --host--toDevice--> opTo(GPU)
                self.apply = self._apply_host_toDevice
            else:
                # case: opFrom(GPU) --toHost--host--toDevice--> opTo(host)
                # Transfers are removed if variables are batched
                if np.any([self.opFrom.discreteFields[v].isBatch
                           for v in self.variables] +
                          [self.opTo.discreteFields[v].isBatch
                           for v in self.variables]):
                    self.apply = self._host
                else:
                    self.apply = self._apply_toHost_host_toDevice

        # Build bridges and toTransfer lists
        self.bridges = {}
        backup = None
        lastvar = None
        # Create bridges between topologies, for each variable.
        for v in self.variables:
            # Bridges creation
            topofrom = self.opFrom.discreteFields[v].topology
            topoto = self.opTo.discreteFields[v].topology
            if backup is not None:
                # Check if a similar bridge has not already been created.
                if [topofrom, topoto] == backup:
                    self.bridges[v] = self.bridges[lastvar]
            else:
                self.bridges[v] = Bridge(topofrom, topoto)
            backup = [topofrom, topoto]
            lastvar = v
            # toTransfer list completion
            if opFrom_is_device:
                self._toHost_fields.append(self.opFrom.discreteFields[v])
            if opTo_is_device:
                self._toDevice_fields.append(self.opTo.discreteFields[v])

        self._main_comm = self.opFrom.discreteFields[v].topology.parent()
        self._main_rank = self._main_comm.Get_rank()

        # Flag telling if there will be some mpi data transfers.
        self._useless_transfer = {}
        for v in self.variables:
            self._useless_transfer[v] = \
                (opFrom_is_device and opTo_is_device) and \
                len(self.bridges[v].recvFrom.keys()) == 0 and \
                len(self.bridges[v].sendTo.keys()) == 0

        # Build MPI subarrays
        dim = self.domain.dimension
        for v in self.variables:
            br = self.bridges[v]
            vToShape = self.opTo.discreteFields[v].data[0].shape
            vFromShape = self.opFrom.discreteFields[v].data[0].shape
            for rk in br.recvFrom.keys():
                subvshape = tuple([br.recvFrom[rk][i].stop -
                                   br.recvFrom[rk][i].start
                                   for i in range(dim)])
                substart = tuple([br.recvFrom[rk][i].start
                                  for i in range(dim)])
                self._r_types[v][rk] = \
                    PARMES_MPI_REAL.Create_subarray(vToShape,
                                                    subvshape,
                                                    substart,
                                                    order=ORDERMPI)
                self._r_types[v][rk].Commit()
            for rk in br.sendTo.keys():
                subvshape = tuple([br.sendTo[rk][i].stop -
                                   br.sendTo[rk][i].start
                                   for i in range(dim)])
                substart = tuple([br.sendTo[rk][i].start
                                  for i in range(dim)])
                self._s_types[v][rk] = \
                    PARMES_MPI_REAL.Create_subarray(vFromShape,
                                                    subvshape,
                                                    substart,
                                                    order=ORDERMPI)
                self._s_types[v][rk].Commit()

        self._isUpToDate = True

    def _apply_toHost_host_toDevice(self, simulation=None):
        if __VERBOSE__:
            print ("{0} APPLY toHOST+HOST+toDEVICE".format(self._main_rank))
        self._toHost()
        self._wait_device()
        self._host()
        self._wait_host()
        self._toDevice()

    def _apply_toHost_host(self, simulation=None):
        if __VERBOSE__:
            print ("{0} APPLY toHOST+HOST".format(self._main_rank))
        self._toHost()
        self._wait_device()
        self._host()

    def _apply_host_toDevice(self, simulation=None):
        if __VERBOSE__:
            print ("{0} APPLY HOST+toDEVICE".format(self._main_rank))
        self._host()
        self._wait_host()
        self._toDevice()

    def _toHost(self):
        """
        Proceed with data transfer of variables from device to host
        """
        for v in self.variables:
            dv = self.opFrom.discreteFields[v]
            if dv in self._toHost_fields:
                if not self._useless_transfer[v]:
                    dv.toHost(self.component)

    def _toDevice(self):
        """
        Proceed with data transfer of variables from device to host
        """
        for v in self.variables:
            dv = self.opTo.discreteFields[v]
            if dv in self._toDevice_fields:
                if not self._useless_transfer[v]:
                    dv.toDevice(self.component)

    def _host(self, simulation=None):
        """
        Proceed with data redistribution from opFrom to opTo
        """
        # TODO :
        # - save a set of bridges in the domain and access them from operator
        # - process all variables in one shot if they have the same topo
        # (use buffers for mpi send/recv? )
        # - move MPI datatypes into the bridge? --> and free MPI type properly
        if __VERBOSE__:
            print ("{0} APPLY HOST".format(self._main_rank))
        self.r_request = {}
        self.s_request = {}
        for v in self.variables:
            br = self.bridges[v]
            # Apply for each component considered
            for d in self._range_components(v):
                if __VERBOSE__:
                    print ("{0} APPLY HOST".format(self._main_rank),
                           self.opFrom.discreteFields[v].name, '->',
                           self.opTo.discreteFields[v].name, S_DIR[d])
                vTo = self.opTo.discreteFields[v].data[d]
                vFrom = self.opFrom.discreteFields[v].data[d]
                v_name = self.opFrom.discreteFields[v].name + S_DIR[d]
                if br.hasLocalInter:
                    vTo[br.ito] = vFrom[br.ifrom]
                cRk = self._main_comm.Get_rank()
                for rk in br.recvFrom.keys():
                    recvtag = (cRk + 1) * 989 + (rk + 1) * 99 + (d + 1) * 88
                    self.r_request[v_name + str(rk)] = \
                        self._main_comm.Irecv([vTo, 1, self._r_types[v][rk]],
                                              source=rk, tag=recvtag)
                    self._hasRequests = True
                for rk in br.sendTo.keys():
                    sendtag = (rk + 1) * 989 + (cRk + 1) * 99 + (d + 1) * 88
                    self.s_request[v_name + str(rk)] = \
                        self._main_comm.Issend([vFrom, 1,
                                               self._s_types[v][rk]],
                                               dest=rk, tag=sendtag)
                    self._hasRequests = True

    def _wait_host(self):
        """
        MPI Barrier to wait for the end
        of all communication requests.
        """
        if __VERBOSE__:
            print ("{0}", "WAIT MPI".format(self._main_rank),
                   self._hasRequests)
        if self._hasRequests:
            for rk in self.r_request.keys():
                self.r_request[rk].Wait()
            for rk in self.s_request.keys():
                self.s_request[rk].Wait()
        self._hasRequests = False

    def _wait_device(self):
        if __VERBOSE__:
            print ("{0}".format(self._main_rank), "WAITING OPENCL")
        for dv in self._toDevice_fields + self._toHost_fields:
            dv.wait()

    def _wait_all(self):
        self._wait_host()
        self._wait_device()

    def test(self, rsend=None, rrecv=None):
        """
        if neither rsend or rrecv is given return
        True if all communication request are complete
        else check for sending to rsend or
        receiving from rrecv. Process ranks
        should be given in main_comm.
        @param rsend : discrete variable name + S_DIR + rank of the process
        to which a message has been sent
        and for which we want to test
        message completion.
        @param  rrecv : discrete variable name + S_DIR + rank of the process
        from which a message has been receive
        and for which we want to test
        message completion.
        """
        if(rsend is not None or rrecv is not None):
            send_res = True
            recv_res = True
            if rsend is not None:
                send_res = self.s_request[rsend].Test()
            if rrecv is not None:
                recv_res = self.r_request[rrecv].Test()
            res = send_res and recv_res
        else:
            res = True
            for rk in self.r_request.keys():
                res = self.r_request[rk].Test()
                if not res:
                    return res
            for rk in self.s_request.keys():
                res = self.s_request[rk].Test()
                if not res:
                    return res
        return res

    def addRedistributeRequirement(self, red):
        raise ValueError(
            "Cannot add a requirement to a Redistribute operator.")

    def getRedistributeRequirement(self):
        return []
