# -*- coding: utf-8 -*-
import parmepy as pp
import numpy as np
from parmepy.fields.continuous import Field
from parmepy.operator.stretching import Stretching
from parmepy.problem.simulation import Simulation
from parmepy.methods_keys import TimeIntegrator, Formulation,\
    SpaceDiscretisation
from parmepy.methods import RK3, FD_C_4, Conservative


def computeVel(res, x, y, z, t):
    amodul = np.cos(np.pi * 1. / 3.)
    pix = np.pi * x
    piy = np.pi * y
    piz = np.pi * z
    pi2x = 2. * pix
    pi2y = 2. * piy
    pi2z = 2. * piz
    res[0][...] = 2. * np.sin(pix) * np.sin(pix) \
        * np.sin(pi2y) * np.sin(pi2z) * amodul
    res[1][...] = - np.sin(pi2x) * np.sin(piy) \
        * np.sin(piy) * np.sin(pi2z) * amodul
    res[2][...] = - np.sin(pi2x) * np.sin(piz) \
        * np.sin(piz) * np.sin(pi2y) * amodul
    return res


def computeVort(res, x, y, z, t):
    amodul = np.cos(np.pi * 1. / 3.)
    pix = np.pi * x
    piy = np.pi * y
    piz = np.pi * z
    pi2x = 2. * pix
    pi2y = 2. * piy
    pi2z = 2. * piz
    res[0][...] = 2. * np.pi * np.sin(pi2x) * amodul *\
        (- np.cos(pi2y) * np.sin(piz) * np.sin(piz)
         + np.sin(piy) * np.sin(piy) * np.cos(pi2z))

    res[1][...] = 2. * np.pi * np.sin(pi2y) * amodul *\
        (2. * np.cos(pi2z) * np.sin(pix) * np.sin(pix)
         + np.sin(piz) * np.sin(piz) * np.cos(pi2x))

    res[2][...] = -2. * np.pi * np.sin(pi2z) * amodul *\
        (np.cos(pi2x) * np.sin(piy) * np.sin(piy)
         + np.sin(pix) * np.sin(pix) * np.cos(pi2y))

    return res


def test_Stretching():
    # Parameters
    nb = 33
    dim = 3
    boxLength = [1., 1., 1.]
    boxMin = [0., 0., 0.]
    nbElem = [nb, nb, nb]

    timeStep = 0.05

    ## Domain
    box = pp.Box(dim, length=boxLength, origin=boxMin)

    ## Fields
    velo = Field(
        domain=box, formula=computeVel,
        name='Velocity', isVector=True)
    vorti = Field(
        domain=box, formula=computeVort,
        name='Vorticity', isVector=True)

    ## Operators
    method = {TimeIntegrator: RK3, Formulation: Conservative,
              SpaceDiscretisation: FD_C_4}
    stretch = Stretching(velo, vorti,
                         resolutions={velo: nbElem,
                                      vorti: nbElem},
                         method=method)
    stretch.discretize()
    topo = stretch.discreteFields[velo].topology
    velo.initialize(topo=topo)
    vorti.initialize(topo=topo)
    stretch.setUp()
    simulation = Simulation(tinit=0, tend=20, timeStep=timeStep)
    stretch.apply(simulation)


if __name__ == "__main__":
    test_Stretching()
