"""
@file parmepy.operator.tests.test_differential
Tests for differential operators.
"""
import numpy as np
from parmepy.domain.box import Box
from parmepy.fields.continuous import Field
#from parmepy.numerics.differential_operations import DivT, GradVxW
#import math as m

# Domain and topologies definitions
nb = 65
Lx = Ly = Lz = 2. * np.pi
box = Box(dimension=3, length=[Lx, Ly, Lz],
          origin=[0., 0., 0.])
resol = [nb, nb, nb]

box2 = Box(dimension=3, length=[2. * Lx, Ly, Lz], origin=[0., 0., 0.])
resol2 = [129, nb, nb]


def callOp(DiffOperator, ref_formula, op_dim=box.dimension, method=None,
           order=4, dom=box, resolution=None):
    """Basic test for Grad operator using FD scheme
    (comparison with periodic analytical solution)
    """
    # Velocity and result fields
    velo = Field(domain=dom, formula=velocity_f, isVector=True)
    result = Field(domain=dom, nbComponents=op_dim)
    if resolution is None:
        resolution = resol
    # Curl operator
    Op = DiffOperator(velo, result, resolutions={velo: resolution,
                                                 result: resolution},
                      method=method)

    Op.discretize()
    topo = Op.discreteFields[velo].topology

    # Reference field
    ref = Field(domain=dom, formula=ref_formula, nbComponents=op_dim)
    ref_d = ref.discretize(topo)
    ref.initialize(topo=topo)
    res_d = result.discreteFields[topo]

    # initialize velocity ...
    velo.initialize(topo=topo)
    Op.setUp()

    # Compute curl
    Op.apply()

    # Compare results with reference
    ind = topo.mesh.iCompute
    err = np.zeros((dom.dimension), dtype=np.float64)
    for i in xrange(dom.dimension):
        err[i] = (dom.length[i] / (resolution[i] - 1)) ** order
    print ('==============')
    print (str(DiffOperator) + ' test')
    print ('==============')
    for i in xrange(result.nbComponents):
        print ('err = O(h**order) =', err[i % dom.dimension])
        print (np.max(np.abs(res_d[i][ind] - ref_d[i][ind])))
        assert np.allclose(res_d[i][ind], ref_d[i][ind],
                           rtol=err[i % dom.dimension])
    #Op.finalize()


def velocity_f(res, x, y, z, t):
    res[0][...] = np.sin(x) * np.cos(y) * np.cos(z)
    res[1][...] = - np.cos(x) * np.sin(y) * np.cos(z)
    res[2][...] = 0.
    return res


def vorticity_f(res, x, y, z, t):
    res[0][...] = - np.cos(x) * np.sin(y) * np.sin(z)
    res[1][...] = - np.sin(x) * np.cos(y) * np.sin(z)
    res[2][...] = 2. * np.sin(x) * np.sin(y) * np.cos(z)
    return res


def grad_velo(res, x, y, z, t):
    res[0][...] = np.cos(x) * np.cos(y) * np.cos(z)
    res[1][...] = -np.sin(x) * np.sin(y) * np.cos(z)
    res[2][...] = -np.sin(x) * np.cos(y) * np.sin(z)
    res[3][...] = np.sin(x) * np.sin(y) * np.cos(z)
    res[4][...] = - np.cos(x) * np.cos(y) * np.cos(z)
    res[5][...] = np.cos(x) * np.sin(y) * np.sin(z)
    res[6][...] = 0.0
    res[7][...] = 0.0
    res[8][...] = 0.0
    return res


def test_CurlFD():
    from parmepy.methods_keys import SpaceDiscretisation
    from parmepy.methods import FD_C_4
    from parmepy.operator.differential import Curl
    method = {SpaceDiscretisation: FD_C_4}
    callOp(Curl, vorticity_f, method=method, resolution=resol)


def test_CurlFD2():
    from parmepy.methods_keys import SpaceDiscretisation
    from parmepy.methods import FD_C_2
    from parmepy.operator.differential import Curl
    method = {SpaceDiscretisation: FD_C_2}
    callOp(Curl, vorticity_f, method=method, order=2, resolution=resol)


def test_CurlFFT():
    from parmepy.methods_keys import SpaceDiscretisation
    from parmepy.operator.differential import Curl
    method = {SpaceDiscretisation: 'fftw'}
    callOp(Curl, vorticity_f, method=method, order=6, resolution=resol)


def test_Grad():
    from parmepy.methods_keys import SpaceDiscretisation
    from parmepy.methods import FD_C_2
    method = {SpaceDiscretisation: FD_C_2}
    from parmepy.operator.differential import Grad
    callOp(Grad, grad_velo, op_dim=9, method=method, order=2, resolution=resol)


def test_Grad2():
    from parmepy.operator.differential import Grad
    from parmepy.methods_keys import SpaceDiscretisation
    from parmepy.methods import FD_C_4
    method = {SpaceDiscretisation: FD_C_4}
    callOp(Grad, grad_velo, op_dim=9, method=method, resolution=resol)


def test_CurlFD_2():
    from parmepy.methods_keys import SpaceDiscretisation
    from parmepy.methods import FD_C_4
    from parmepy.operator.differential import Curl
    method = {SpaceDiscretisation: FD_C_4}
    callOp(Curl, vorticity_f, method=method, dom=box2, resolution=resol2)


def test_CurlFD2_2():
    from parmepy.methods_keys import SpaceDiscretisation
    from parmepy.methods import FD_C_2
    from parmepy.operator.differential import Curl
    method = {SpaceDiscretisation: FD_C_2}
    callOp(Curl, vorticity_f, method=method, order=2, dom=box2,
           resolution=resol2)


def test_CurlFFT_2():
    from parmepy.methods_keys import SpaceDiscretisation
    from parmepy.operator.differential import Curl
    method = {SpaceDiscretisation: 'fftw'}
    callOp(Curl, vorticity_f, method=method, order=6, dom=box2,
           resolution=resol2)


def test_Grad_2():
    from parmepy.methods_keys import SpaceDiscretisation
    from parmepy.methods import FD_C_2
    method = {SpaceDiscretisation: FD_C_2}
    from parmepy.operator.differential import Grad
    callOp(Grad, grad_velo, op_dim=9, method=method, order=2, dom=box2,
           resolution=resol2)


def test_Grad2_2():
    from parmepy.operator.differential import Grad
    from parmepy.methods_keys import SpaceDiscretisation
    from parmepy.methods import FD_C_4
    method = {SpaceDiscretisation: FD_C_4}
    callOp(Grad, grad_velo, op_dim=9, method=method, dom=box2,
           resolution=resol2)


# This may be useful to run mpi tests
if __name__ == "__main__":
    test_CurlFD()
    test_CurlFD2()
    test_CurlFFT()
    test_Grad()
    test_Grad2()
    test_CurlFD_2()
    test_CurlFD2_2()
    test_CurlFFT_2()
    test_Grad_2()
    test_Grad2_2()
