# -*- coding: utf-8 -*-
"""
Tests for reader/writer of fields in hdf5 format.
"""

from parmepy.operator.monitors.printer import Printer
from parmepy.operator.monitors.reader import Reader
from parmepy import Box, Field
from parmepy.mpi.topology import Cartesian
import numpy as np
import os
from parmepy.constants import HDF5
from parmepy.problem.simulation import Simulation
import shutil

Lx = Ly = Lz = 2.
nb = 65
workingDir = os.getcwd() + '/test_hdf5/'


def init(dim):
    # Domain (box-shaped)
    dom = Box(dimension=dim, length=[Lx]*dim, origin=[-1.]*dim)
    # global resolution for the grid
    resol = [nb] * dim
    ghosts = np.zeros(dim) + 2
    topo = Cartesian(dom, dom.dimension, resol, ghosts=ghosts)

    return dom, topo


# Example of function to init the scalar field
def func3D(res, x, y, z, t):
    res[0][...] = np.cos(t * x) + np.sin(y) + z
    return res


def purgeFiles():
    shutil.rmtree(workingDir)


def test_write_read_scalar_3D():
    dom, topo = init(3)
    scal3D = Field(domain=dom, name='Scal3D')
    scalRef = Field(domain=dom, formula=func3D, name='ScalRef3D')
    scRef = scalRef.discretize(topo)
    sc3D = scal3D.discretize(topo)
    printer = Printer(variables=[scalRef], topo=topo,
                      formattype=HDF5, prefix=workingDir+'/testIO_scal')
    simu = Simulation(tinit=0.0, tend=2., nbIter=100, iterMax=1000000)
    scalRef.initialize(currentTime=simu.start, topo=topo)
    # Print scalRef for t = tinit.
    printer.apply(simu)
    simu.initialize()
    simu.advance()
    simu.advance()
    # Print scalRef for another time
    printer.apply(simu)
    printer.finalize()
    assert os.path.exists(workingDir + '/testIO_scal.xmf')
    assert os.path.exists(workingDir + '/testIO_scal_00000.h5')
    assert os.path.exists(workingDir + '/testIO_scal_00001.h5')
    # Reader
    reader = Reader(variables=[scal3D], topo=topo,
                    prefix=workingDir+'/testIO_scal_00000')
  #  print reader.dataset_names()
  #  print reader.names
    assert not np.allclose(sc3D.data[0][topo.mesh.iCompute],
                           scRef.data[0][topo.mesh.iCompute])
    reader.apply()
    reader.finalize()
    # Check if scal3D is equal to scalRef
    assert np.allclose(sc3D.data[0][topo.mesh.iCompute],
                       scRef.data[0][topo.mesh.iCompute])
    purgeFiles()


def test_write_read_2scalar_3D():
    dom, topo = init(3)
    scal3D = Field(domain=dom, name='Scal3D')
    scalRef = Field(domain=dom, formula=func3D, name='ScalRef3D')
    scRef = scalRef.discretize(topo)
    sc3D = scal3D.discretize(topo)
    scal3D2 = Field(domain=dom, name='Scal3D_2')
    scalRef2 = Field(domain=dom, formula=func3D, name='ScalRef23D')
    scRef2 = scalRef2.discretize(topo)
    sc3D2 = scal3D2.discretize(topo)
    printer = Printer(variables=[scalRef, scalRef2], topo=topo,
                      formattype=HDF5, prefix=workingDir+'/testIO_scal')
    simu = Simulation(tinit=0.0, tend=2., nbIter=100, iterMax=1000000)
    scalRef.initialize(currentTime=simu.start, topo=topo)
    scalRef2.initialize(currentTime=simu.start, topo=topo)
    scRef2[0][...] *= 3.
    printer.apply(simu)
    printer.finalize()
    # Print scalRef/ScalRef2 for t = tinit.
    # Read only scalRef2 into scal3D
    reader = Reader(variables=[scal3D], topo=topo,
                    prefix=workingDir+'/testIO_scal_00000',
                    names={scal3D: 'ScalRef23D'})
    assert not np.allclose(sc3D.data[0][topo.mesh.iCompute],
                           scRef2.data[0][topo.mesh.iCompute])
    reader.apply()
    reader.finalize()
    # Check if scal3D is equal to scalRef
    assert np.allclose(sc3D.data[0][topo.mesh.iCompute],
                       scRef2.data[0][topo.mesh.iCompute])
    sc3D[0][...] = 0.0
    # Read scalRef and scalRef2 into scal3D and scal3D2
    reader2 = Reader(variables=[scal3D, scal3D2], topo=topo,
                     prefix=workingDir+'/testIO_scal_00000',
                     names={scal3D: 'ScalRef3D', scal3D2: 'ScalRef23D'})
    #    assert not np.allclose(sc3D.data[0][topo.mesh.iCompute],
    #                       scRef2.data[0][topo.mesh.iCompute])
    reader2.apply()
    reader2.finalize()
    # Check if scal3D is equal to scalRef and scal3D2 == scalRef2
    assert np.allclose(sc3D.data[0][topo.mesh.iCompute],
                       scRef.data[0][topo.mesh.iCompute])
    assert np.allclose(sc3D2.data[0][topo.mesh.iCompute],
                       scRef2.data[0][topo.mesh.iCompute])
    purgeFiles()
