# -*- coding: utf-8 -*-
"""
@file operator/velocity_correction.py

Operator to shift velocity to fit with a required input flowrate.

"""
from parmepy.constants import debug
from parmepy.operator.discrete.velocity_correction import VelocityCorrection_D
from parmepy.operator.continuous import Operator
from parmepy.domain.obstacle.controlBox import ControlBox


class VelocityCorrection(Operator):
    """
    The velocity field is corrected after solving the
    Poisson equation. For more details about calculations,
    see the "velocity_correction.pdf" explanations document
    in ParmeDoc directory.
    """

    @debug
    def __init__(self, velocity, vorticity, req_flowrate, surf=None, **kwds):
        """
        Corrects the values of the velocity field after
        solving Poisson equation in order to prescribe proper
        mean flow and ensure the desired inlet flowrate.

        @param[in, out] velocity field to be corrected
        @param[in] vorticity field used to compute correction
        @param resolutions : grid resolutions of velocity and vorticity
        @param[in] req_flowrate : required value for the flowrate
        @param[in] surf : surface (parmepy.domain.obstacle.planes.SubPlane)
        used to compute reference flow rates. Default = surface at x_origin,
        normal to x-dir.
        @param topo : a predefined topology to discretize velocity/vorticity
        """
        super(VelocityCorrection, self).__init__(variables=[velocity,
                                                            vorticity], **kwds)
        ## velocity variable (vector)
        self.velocity = velocity
        ## velocity variable
        self.vorticity = vorticity

        self.input = [self.velocity, self.vorticity]
        self.output = [self.velocity]
        ## A 'reference' surface on which flow rates will be computed
        ## (usually, surface normal to the flow at origin)
        self.surfRef = surf
        ## Expected value for the flow rate through self.surfRef
        self.req_flowrate = req_flowrate
        dom = self.velocity.domain
        self.cb = ControlBox(domain=dom, origin=dom.origin, lengths=dom.length)

    def discretize(self):
        super(VelocityCorrection, self)._standard_discretize()

    @debug
    def setUp(self):
        self.discreteOperator =\
            VelocityCorrection_D(self.discreteFields[self.velocity],
                                 self.discreteFields[self.vorticity],
                                 self.req_flowrate, self.cb)

        self.discreteOperator.setUp()
        self._isUpToDate = True

    def apply(self, simulation=None):
        """
        Compute correction and add it to current velocoty
        """
        self.discreteOperator.apply(simulation)

    def computeCorrection(self):
        """
        Compute the required correction for the current state
        but do not apply it onto velocity.
        """
        self.discreteOperator.computeCorrection()
