"""
@file problem.py

Complete problem description.
"""
from parmepy.constants import debug, parmesPickle
from parmepy import __VERBOSE__
from parmepy.operator.monitors.monitoring import Monitoring
from parmepy.operator.redistribute import Redistribute
from parmepy.tools.timers import Timer, timed_function
from parmepy.mpi import main_rank


class Problem(object):
    """
    Problem representation.

    Contains several operators and monitors that apply on variables.
    Variables are defined on different domains.\n
    Each operator is set up and variables are initialized in a set up step.\n
    To solve the problem, a loop over time-steps is launched. A step consists
    in calling the apply method of each operators and monitors.\n
    To finish, a finalize method is called.\
    """

    @debug
    def __new__(cls, *args, **kw):
        return object.__new__(cls, *args, **kw)

    @debug
    def __init__(self, operators, simulation, monitors=None,
                 dumpFreq=100, name=None):
        """
        Create a transport problem instance.

        @param operators : list of operators.
        @param simulation : a parmepy.simulation.Simulation object
        to describe simulation parameters.
        @param monitors : list of monitors.
        @param name : an id for the problem
        @param dumpFreq : frequency of dump (i.e. saving to a file)
        for the problem; set dumpFreq = -1 for no dumps. Default = 100.
        """
        ## Problem name
        self.name = name
        ## Problem operators
        self.operators = operators
        if monitors is None:
            monitors = []
        for m in monitors:
            self.operators.append(m)
        ## Computes time step and manage iterations
        self.simulation = simulation
        self.domain = self.operators[0].variables[0].domain
        for op in self.operators:
            for v in op.variables:
                if not self.domain is v.domain:
                    raise ValueError("Problem must have only one " +
                                     "domain for variables.")
        ## A list of variables that must be initialized before
        ## any call to op.apply()
        self.input = []
        ## call to problem.dump frequency during apply.
        if dumpFreq >= 0:
            ## dump problem every self.dumpFreq iter
            self.dumpFreq = dumpFreq
            self._doDump = True
        else:
            self._doDump = False
            self.dumpFreq = 100000

        ## Id for the problem. Used for dump file name.
        if name is None:
            self.name = 'parmesPb'
        else:
            self.name = name
        ## Object to store computational times of lower level functions
        self.timer = Timer(self)
        ## Default file name prefix for dump.
        self.filename = str(self.name)
        self._filedump = self.filename + '_rk_' + str(main_rank)

        # Flag : true when operators for computation are up
        # and when variables are initialized (i.e. after a call to pre_setUp)
        # Note : 3 categories of op : computation (stretching, poisson ...),
        # monitors (printer ...) and data distribution (Redistribute)
        self._isReady = False

    def addMonitors(self, monitors, position=None):
        """
        Set monitors for the problem
        If position is given, insert monitor[i] at position[i]
        else, monitior are added at the end of the list.
        """
        if position is None:
            for m in monitors:
                self.operators.append(m)
        else:
            for m, pos in zip(monitors, position):
                self.operators.insert(pos, m)

    @debug
    @timed_function
    def setUp(self):
        """
        Prepare operators (create topologies, allocate memories ...)
        """
        # Set up for 'computational' operators
        if not self._isReady:
            self.pre_setUp()

        for v in self.input:
            v.initialize()

        # other operators
        for op in self.operators:
            if isinstance(op, Monitoring):
                op.setUp()
                op.discretize()

        for op in self.operators:
            if isinstance(op, Redistribute):
                op.setUp()

        if __VERBOSE__ and main_rank == 0:
            print ("====")

    def pre_setUp(self):
        """
        - Partial setup : only for 'computational' operators
        (i.e. excluding monitors, rendering, data distribution ...)
        - Initialize variables.
        """
        if self._isReady:
            pass

        for op in self.operators:
            if not isinstance(op, Monitoring):
                if not isinstance(op, Redistribute):
                    op.discretize()

        for op in self.operators:
            if not isinstance(op, Monitoring):
                if not isinstance(op, Redistribute):
                    op.setUp()

        if __VERBOSE__ and main_rank == 0:
            print ("==== Variables initialization ====")

        # Build variables list to initialize
        # These are operators input variables that are not output of
        # previous operators in the operator stack.
        # Set the variables input topology as the the topology of the fist
        # operator that uses this variable as input.
        self.input = []
        for op in self.operators[::-1]:
            for v in op.output:
                if v in self.input:
                    self.input.remove(v)
            for v in op.input:
                if not v in self.input:
                    self.input.append(v)
#                    if not isinstance(op, Monitoring):
#                        if not isinstance(op, Redistribute):
#                            v.setTopoInit(op.discreteFields[v].topology)

        self._isReady = True

    @debug
    @timed_function
    def solve(self):
        """
        Solve problem.

        Performs simulations iterations by calling each
        operators of the list until timer ends.\n
        At end of time step, call an io step.\n
        Displays timings at simulation end.
        """
        self.simulation.initialize()
        if main_rank == 0:
            print ("\n\n Start solving ...")
        while not self.simulation.isOver:
            if main_rank == 0:
                self.simulation.printState()

            for op in self.operators:
                if __VERBOSE__:
                    print (main_rank, op.name)
                op.apply(self.simulation)

            testdump = \
                self.simulation.currentIteration % self.dumpFreq is 0
            self.simulation.advance()
            if self._doDump and testdump:
                self.dump()

    @debug
    def finalize(self):
        """
        Finalize method
        """
        if main_rank == 0:
            print ("\n\n==== End ====")
        ## We terminate monitors before operators.
        for op in self.operators:
            if isinstance(op, Monitoring):
                op.finalize()
                self.timer = self.timer + op.timer
        for op in self.operators:
            if not isinstance(op, Monitoring):
                op.finalize()
                self.timer = self.timer + op.timer
        var = []
        for op in self.operators:
            for v in op.variables:
                if not v in var:
                    var.append(v)
        for v in var:
            v.finalize()
            try:
                self.timer = self.timer + v.timer
            except AttributeError:
                pass
        if main_rank == 0:
            print ("===\n")

    def __str__(self):
        """ToString method"""
        s = "Problem based on\n"
        s += str(self.domain.topologies)
        s += "with following operators : \n"
        for op in self.operators:
            s += str(op)
        return s

    def dump(self, filename=None):
        """
        Serialize some data of the problem to file
        (only data required for a proper restart, namely fields in self.input
        and simulation).
        @param filename : prefix for output file. Real name = filename_rk_N,
        N being current process number. If None use default value from problem
        parameters (self.filename)
        """
        if filename is not None:
            self.filename = filename
            self._filedump = filename + '_rk_' + str(main_rank)
        db = parmesPickle(self._filedump, mode='store')
        db.dump(self.simulation, 'simulation')
        for v in self.input:
            v.dump(self.filename, mode='append')

    def restart(self, filename=None):
        """
        Load serialized data to restart from a previous state.
        self.input variables and simulation are loaded.
        @param  filename : prefix for downloaded file.
        Real name = filename_rk_N, N being current process number.
        If None use default value from problem
        parameters (self.filename)
        """
        if filename is not None:
            self.filename = filename
            self._filedump = filename + '_rk_' + str(main_rank)
        db = parmesPickle(self._filedump, mode='load')
        self.simulation = db.load('simulation')[0]
        self.simulation.reset()
        for v in self.input:
            print ("load ...", self.filename)
            v.load(self.filename)

        for op in self.operators:
            if isinstance(op, Monitoring):
                op.setUp()
            if isinstance(op, Redistribute):
                op.setUp()

    def setDumpFreq(self, freq):
        """
        set rate of problem.dump call (every 'rate' iteration)
        @param freq : the frequency of output
        """
        self.dumpFreq = freq
        if freq < 0:
            self._doDump = False
