"""
@file problem_tasks.py

Extending problem description to handle tasks parallelism.
Each operator owns a task id that define a process group that are sharing the
same tasks.
"""
from parmepy.constants import debug
from parmepy import __VERBOSE__
from parmepy.problem.problem import Problem
from parmepy.operator.monitors.monitoring import Monitoring
from parmepy.operator.redistribute import Redistribute
from parmepy.operator.redistribute_intercomm import RedistributeIntercomm
from parmepy.tools.timers import timed_function


class ProblemTasks(Problem):
    """
    As in Problem, it contains several operators and monitors that apply
    on variables. The operators are labeled by task_id that defines
    a identifier of a task.
    Tasks are subset of operators and are assigned to a subset of the MPI
    process by means of the task_list parameter.
    """
    @debug
    def __new__(cls, *args, **kw):
        return object.__new__(cls, *args, **kw)

    @debug
    def __init__(self, operators, simulation, tasks_list, monitors=None,
                 dumpFreq=100, name=None, main_comm=None):
        """
        Creates the problem.
        @param operators : list of operators.
        @param simulation : a parmepy.simulation.Simulation object
        to describe simulation parameters.
        @param tasks_list : list of task identifiers for each process rank
        @param monitors : list of monitors.
        @param name : an id for the problem
        @param dumpFreq : frequency of dump (i.e. saving to a file)
        for the problem; set dumpFreq = -1 for no dumps. Default = 100.
        @param main_comm : MPI communicator that contains all process
        involved in this problem.

        @remark : process number in communicator main_comm must equal the
        length of tasks_list.
        """
        Problem.__init__(self, operators, simulation, monitors=monitors,
                         dumpFreq=dumpFreq, name=name)
        self.tasks_list = tasks_list
        if main_comm is None:
            from parmepy.mpi.main_var import main_comm
        self.main_comm = main_comm
        self._main_rank = self.main_comm.Get_rank()
        assert self.main_comm.Get_size() == len(self.tasks_list), \
            "The given task list length (" + str(self.tasks_list) + ") " \
            "does not match the communicator size" \
            " ({0})".format(self.main_comm.Get_size())

    def pre_setUp(self):
        """
        - Removes operators and monitors that not have the same task identifier
        as the current process
        - Keep the Redistribute_intercomm in both 'from' and 'to' task_id
        - Partial setup : only for 'computational' operators
        (i.e. excluding monitors, rendering, data distribution ...)
        - Initialize variables.
        """
        if self._isReady:
            pass

        self.operators_backup = []
        ## Remove operators with a tasks not handled by this process.
        for op in self.operators[::-1]:
            self.operators_backup.append(op)
            if not isinstance(op, RedistributeIntercomm):
                if op.task_id != self.tasks_list[self._main_rank]:
                    self.operators.remove(op)
            else:
                if op.id_from != self.tasks_list[self._main_rank] and \
                        op.id_to != self.tasks_list[self._main_rank]:
                    self.operators.remove(op)

        # Discretize and setup computational operators
        for op in self.operators:
            if not isinstance(op, Monitoring):
                if not isinstance(op, Redistribute):
                    if not isinstance(op, RedistributeIntercomm):
                        op.discretize()
        for op in self.operators:
            if not isinstance(op, Monitoring):
                if not isinstance(op, Redistribute):
                    if not isinstance(op, RedistributeIntercomm):
                        op.setUp()

        # Build variables list to initialize
        # These are operators input variables that are not output of
        # previous operators in the operator stack.
        # Set the variables input topology as the the topology of the fist
        # operator that uses this variable as input.
        self.input = []
        for op in self.operators[::-1]:
            for v in op.output:
                if v in self.input:
                    self.input.remove(v)
            for v in op.input:
                if not v in self.input:
                    if isinstance(op, RedistributeIntercomm):
                        if op.id_from == self.tasks_list[self._main_rank]:
                            self.input.append(v)
                    else:
                        self.input.append(v)
#                    if not isinstance(op, Monitoring):
#                        if not isinstance(op, Redistribute):
#                            if not isinstance(op, RedistributeIntercomm):
#                                v.setTopoInit(op.discreteFields[v].topology)

        self._isReady = True

    def addMonitors(self, monitors, position=None):
        """
        Set monitors for the problem
        If position is given, insert monitor[i] at position[i]
        else, monitior are added at the end of the list.
        """
        if position is None:
            for m in monitors:
                if m.task_id == self.tasks_list[self._main_rank]:
                    self.operators.append(m)
        else:
            for m, pos in zip(monitors, position):
                if m.task_id == self.tasks_list[self._main_rank]:
                    self.operators.insert(pos, m)

    @debug
    @timed_function
    def setUp(self):
        """
        Prepare operators (create topologies, allocate memories ...)
        """
        # Set up for 'computational' operators
        if not self._isReady:
            self.pre_setUp()

        for v in self.input:
            v.initialize()

        # other operators
        for op in self.operators:
            if isinstance(op, Monitoring):
                op.setUp()

        for op in self.operators:
            if isinstance(op, RedistributeIntercomm):
                op.discretize()
                op.setUp()

        for op in self.operators:
            if isinstance(op, Redistribute):
                op.setUp()

        if __VERBOSE__ and self._main_rank == 0:
            print ("====")

    @debug
    @timed_function
    def solve(self):
        """
        Solve problem.

        Performs simulations iterations by calling each
        operators of the list until timer ends.\n
        At end of time step, call an io step.\n
        Displays timings at simulation end.
        """
        self.simulation.initialize()
        self.main_comm.Barrier()
        if self._main_rank == 0:
            print ("\n\n Start solving ...")
        while not self.simulation.isOver:
            if self._main_rank == 0:
                self.simulation.printState()
            for op in self.operators:
                op.apply(self.simulation)
            testdump = \
                self.simulation.currentIteration % self.dumpFreq is 0
            self.simulation.advance()
            if self._doDump and testdump:
                self.dump()

    @debug
    def finalize(self):
        """
        Finalize method
        """
        if self._main_rank == 0:
            print ("\n\n==== End ====")
        ## We terminate monitors before operators.
        for op in self.operators:
            if isinstance(op, Monitoring):
                op.finalize()
                self.timer = self.timer + op.timer
        for op in self.operators:
            if not isinstance(op, Monitoring):
                op.finalize()
                self.timer = self.timer + op.timer
        var = []
        for op in self.operators:
            for v in op.variables:
                if not v in var:
                    var.append(v)
        for v in var:
            v.finalize()
            try:
                self.timer = self.timer + v.timer
            except AttributeError:
                pass
        if self._main_rank == 0:
            print ("===\n")
